# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii                 import *
from torii.hdl.rec         import Layout
from torii.sim             import *
from torii.lib.soc.csr.bus import *
from torii.lib.soc.memory  import MemoryMap

from ....utils             import ToriiTestSuiteCase
class ElementTestCase(ToriiTestSuiteCase):
	def test_layout_1_ro(self):
		elem = Element(1, 'r')
		self.assertEqual(elem.width, 1)
		self.assertEqual(elem.access, Element.Access.R)
		self.assertEqual(elem.layout, Layout.cast([
			('r_data', 1),
			('r_stb', 1),
		]))

	def test_layout_8_rw(self):
		elem = Element(8, access = 'rw')
		self.assertEqual(elem.width, 8)
		self.assertEqual(elem.access, Element.Access.RW)
		self.assertEqual(elem.layout, Layout.cast([
			('r_data', 8),
			('r_stb', 1),
			('w_data', 8),
			('w_stb', 1),
		]))

	def test_layout_10_wo(self):
		elem = Element(10, 'w')
		self.assertEqual(elem.width, 10)
		self.assertEqual(elem.access, Element.Access.W)
		self.assertEqual(elem.layout, Layout.cast([
			('w_data', 10),
			('w_stb', 1),
		]))

	def test_layout_0_rw(self): # degenerate but legal case
		elem = Element(0, access = Element.Access.RW)
		self.assertEqual(elem.width, 0)
		self.assertEqual(elem.access, Element.Access.RW)
		self.assertEqual(elem.layout, Layout.cast([
			('r_data', 0),
			('r_stb', 1),
			('w_data', 0),
			('w_stb', 1),
		]))

	def test_width_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Width must be a non-negative integer, not -1'
		):
			Element(-1, 'rw')

	def test_access_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Access mode must be one of \'r\', \'w\', or \'rw\', not \'wo\''
		):
			Element(width = 1, access = 'wo')


class InterfaceTestCase(ToriiTestSuiteCase):
	def test_layout(self):
		iface = Interface(addr_width = 12, data_width = 8)
		self.assertEqual(iface.addr_width, 12)
		self.assertEqual(iface.data_width, 8)
		self.assertEqual(iface.layout, Layout.cast([
			('addr',    12),
			('r_data',  8),
			('r_stb',   1),
			('w_data',  8),
			('w_stb',   1),
		]))

	def test_wrong_addr_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Address width must be a positive integer, not -1'
		):
			Interface(addr_width = -1, data_width = 8)

	def test_wrong_data_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Data width must be a positive integer, not -1'
		):
			Interface(addr_width = 16, data_width = -1)

	def test_get_map_wrong(self):
		iface = Interface(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			NotImplementedError,
			r'Bus interface \(rec iface addr r_data r_stb w_data w_stb\) does not '
			r'have a memory map'
		):
			iface.memory_map

	def test_get_map_frozen(self):
		iface = Interface(addr_width = 16, data_width = 8)
		iface.memory_map = MemoryMap(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen. Address width cannot be extended '
			r'further'
		):
			iface.memory_map.addr_width = 24

	def test_set_map_wrong(self):
		iface = Interface(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Memory map must be an instance of MemoryMap, not \'foo\''
		):
			iface.memory_map = 'foo'

	def test_set_map_wrong_addr_width(self):
		iface = Interface(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has address width 8, which is not the same as '
			r'bus interface address width 16'
		):
			iface.memory_map = MemoryMap(addr_width = 8, data_width = 8)

	def test_set_map_wrong_data_width(self):
		iface = Interface(addr_width = 16, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has data width 16, which is not the same as '
			r'bus interface data width 8'
		):
			iface.memory_map = MemoryMap(addr_width = 16, data_width = 16)


class MultiplexerTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.dut = Multiplexer(addr_width = 16, data_width = 8)

	def test_add_4b(self):
		elem_4b = Element(4, 'rw')
		self.assertEqual(self.dut.add(elem_4b), (0, 1))

	def test_add_8b(self):
		elem_8b = Element(8, 'rw')
		self.assertEqual(self.dut.add(elem_8b), (0, 1))

	def test_add_12b(self):
		elem_12b = Element(12, 'rw')
		self.assertEqual(self.dut.add(elem_12b), (0, 2))

	def test_add_16b(self):
		elem_16b = Element(16, 'rw')
		self.assertEqual(self.dut.add(elem_16b), (0, 2))

	def test_add_two(self):
		elem_8b  = Element( 8, 'rw')
		elem_16b = Element(16, 'rw')
		self.assertEqual(self.dut.add(elem_16b), (0, 2))
		self.assertEqual(self.dut.add(elem_8b),  (2, 3))

	def test_add_extend(self):
		elem_8b  = Element(8, 'rw')
		self.assertEqual(
			self.dut.add(elem_8b, addr = 0x10000, extend = True),
			(0x10000, 0x10001)
		)
		self.assertEqual(self.dut.bus.addr_width, 17)

	def test_add_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Element must be an instance of csr\.Element, not \'foo\''
		):
			self.dut.add(element = 'foo')

	def test_align_to(self):
		elem_0  = Element( 8, 'rw')
		elem_1  = Element( 8, 'rw')
		self.assertEqual(self.dut.add(elem_0), (0, 1))
		self.assertEqual(self.dut.align_to(2), 4)
		self.assertEqual(self.dut.add(elem_1), (4, 5))

	def test_add_wrong_out_of_bounds(self):
		elem = Element(8, 'rw')
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x10000\.\.0x10001 out of bounds for memory map spanning '
			r'range 0x0\.\.0x10000 \(16 address bits\)'
		):
			self.dut.add(elem, addr = 0x10000)

	def test_sim(self):
		elem_4_r = Element(4, 'r')
		self.dut.add(elem_4_r)
		elem_8_w = Element(8, 'w')
		self.dut.add(elem_8_w)
		elem_16_rw = Element(16, 'rw')
		self.dut.add(elem_16_rw)

		bus = self.dut.bus

		def sim_test():
			yield elem_4_r.r_data.eq(0xa)
			yield elem_16_rw.r_data.eq(0x5aa5)

			yield bus.addr.eq(0)
			yield bus.r_stb.eq(1)
			yield
			yield bus.r_stb.eq(0)
			self.assertEqual((yield elem_4_r.r_stb), 1)
			self.assertEqual((yield elem_16_rw.r_stb), 0)
			yield
			self.assertEqual((yield bus.r_data), 0xa)

			yield bus.addr.eq(2)
			yield bus.r_stb.eq(1)
			yield
			yield bus.r_stb.eq(0)
			self.assertEqual((yield elem_4_r.r_stb), 0)
			self.assertEqual((yield elem_16_rw.r_stb), 1)
			yield
			yield bus.addr.eq(3) # pipeline a read
			self.assertEqual((yield bus.r_data), 0xa5)

			yield bus.r_stb.eq(1)
			yield
			yield bus.r_stb.eq(0)
			self.assertEqual((yield elem_4_r.r_stb), 0)
			self.assertEqual((yield elem_16_rw.r_stb), 0)
			yield
			self.assertEqual((yield bus.r_data), 0x5a)

			yield bus.addr.eq(1)
			yield bus.w_data.eq(0x3d)
			yield bus.w_stb.eq(1)
			yield
			yield bus.w_stb.eq(0)
			yield bus.addr.eq(2) # change address
			yield
			self.assertEqual((yield elem_8_w.w_stb), 1)
			self.assertEqual((yield elem_8_w.w_data), 0x3d)
			self.assertEqual((yield elem_16_rw.w_stb), 0)
			yield
			self.assertEqual((yield elem_8_w.w_stb), 0)

			yield bus.addr.eq(2)
			yield bus.w_data.eq(0x55)
			yield bus.w_stb.eq(1)
			yield
			self.assertEqual((yield elem_8_w.w_stb), 0)
			self.assertEqual((yield elem_16_rw.w_stb), 0)
			yield bus.addr.eq(3) # pipeline a write
			yield bus.w_data.eq(0xaa)
			yield
			self.assertEqual((yield elem_8_w.w_stb), 0)
			self.assertEqual((yield elem_16_rw.w_stb), 0)
			yield bus.w_stb.eq(0)
			yield
			self.assertEqual((yield elem_8_w.w_stb), 0)
			self.assertEqual((yield elem_16_rw.w_stb), 1)
			self.assertEqual((yield elem_16_rw.w_data), 0xaa55)

		sim = Simulator(self.dut)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()


class MultiplexerAlignedTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.dut = Multiplexer(addr_width = 16, data_width = 8, alignment = 2)

	def test_add_two(self):
		elem_0 = Element( 8, 'rw')
		elem_1 = Element(16, 'rw')
		self.assertEqual(self.dut.add(elem_0), (0, 4))
		self.assertEqual(self.dut.add(elem_1), (4, 8))

	def test_over_align_to(self):
		elem_0 = Element(8, 'rw')
		elem_1 = Element(8, 'rw')
		self.assertEqual(self.dut.add(elem_0), (0, 4))
		self.assertEqual(self.dut.align_to(3), 8)
		self.assertEqual(self.dut.add(elem_1), (8, 12))

	def test_under_align_to(self):
		elem_0 = Element(8, 'rw')
		elem_1 = Element(8, 'rw')
		self.assertEqual(self.dut.add(elem_0), (0, 4))
		self.assertEqual(self.dut.align_to(alignment = 1), 4)
		self.assertEqual(self.dut.add(elem_1), (4, 8))

	def test_sim(self):
		elem_20_rw = Element(20, 'rw')
		self.dut.add(elem_20_rw)

		bus = self.dut.bus

		def sim_test():
			yield bus.w_stb.eq(1)
			yield bus.addr.eq(0)
			yield bus.w_data.eq(0x55)
			yield
			self.assertEqual((yield elem_20_rw.w_stb), 0)
			yield bus.addr.eq(1)
			yield bus.w_data.eq(0xaa)
			yield
			self.assertEqual((yield elem_20_rw.w_stb), 0)
			yield bus.addr.eq(2)
			yield bus.w_data.eq(0x33)
			yield
			self.assertEqual((yield elem_20_rw.w_stb), 0)
			yield bus.addr.eq(3)
			yield bus.w_data.eq(0xdd)
			yield
			self.assertEqual((yield elem_20_rw.w_stb), 0)
			yield bus.w_stb.eq(0)
			yield
			self.assertEqual((yield elem_20_rw.w_stb), 1)
			self.assertEqual((yield elem_20_rw.w_data), 0x3aa55)

		sim = Simulator(self.dut)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()


class DecoderTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.dut = Decoder(addr_width = 16, data_width = 8)

	def test_align_to(self):
		sub_1 = Interface(addr_width = 10, data_width = 8)
		sub_1.memory_map = MemoryMap(addr_width = 10, data_width = 8)
		self.assertEqual(self.dut.add(sub_1), (0, 0x400, 1))

		self.assertEqual(self.dut.align_to(12), 0x1000)
		self.assertEqual(self.dut.align_to(alignment = 12), 0x1000)

		sub_2 = Interface(addr_width = 10, data_width = 8)
		sub_2.memory_map = MemoryMap(addr_width = 10, data_width = 8)
		self.assertEqual(self.dut.add(sub_2), (0x1000, 0x1400, 1))

	def test_add_extend(self):
		iface = Interface(addr_width = 17, data_width = 8)
		iface.memory_map = MemoryMap(addr_width = 17, data_width = 8)
		self.assertEqual(self.dut.add(iface, extend = True), (0, 0x20000, 1))
		self.assertEqual(self.dut.bus.addr_width, 18)

	def test_add_wrong_sub_bus(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Subordinate bus must be an instance of csr\.Interface, not 1'
		):
			self.dut.add(sub_bus = 1)

	def test_add_wrong_data_width(self):
		mux = Multiplexer(addr_width = 10, data_width = 16)
		Fragment.get(mux, platform = None) # silence UnusedElaboratable

		with self.assertRaisesRegex(
			ValueError,
			r'Subordinate bus has data width 16, which is not the same as '
			r'decoder data width 8'
		):
			self.dut.add(mux.bus)

	def test_add_wrong_out_of_bounds(self):
		iface = Interface(addr_width = 17, data_width = 8)
		iface.memory_map = MemoryMap(addr_width = 17, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x0\.\.0x20000 out of bounds for memory map spanning '
			r'range 0x0\.\.0x10000 \(16 address bits\)'
		):
			self.dut.add(iface)

	def test_sim(self):
		mux_1  = Multiplexer(addr_width = 10, data_width = 8)
		elem_1 = Element(8, 'rw')
		mux_1.add(elem_1)
		self.dut.add(mux_1.bus)

		mux_2  = Multiplexer(addr_width = 10, data_width = 8)
		elem_2 = Element(8, 'rw')
		mux_2.add(elem_2, addr = 2)
		self.dut.add(mux_2.bus)

		elem_1_info = self.dut.bus.memory_map.find_resource(elem_1)
		elem_2_info = self.dut.bus.memory_map.find_resource(elem_2)
		elem_1_addr = elem_1_info.start
		elem_2_addr = elem_2_info.start
		self.assertEqual(elem_1_addr, 0x0000)
		self.assertEqual(elem_2_addr, 0x0402)

		bus = self.dut.bus

		def sim_test():
			yield bus.addr.eq(elem_1_addr)
			yield bus.w_stb.eq(1)
			yield bus.w_data.eq(0x55)
			yield
			yield bus.w_stb.eq(0)
			yield
			self.assertEqual((yield elem_1.w_data), 0x55)

			yield bus.addr.eq(elem_2_addr)
			yield bus.w_stb.eq(1)
			yield bus.w_data.eq(0xaa)
			yield
			yield bus.w_stb.eq(0)
			yield
			self.assertEqual((yield elem_2.w_data), 0xaa)

			yield elem_1.r_data.eq(0x55)
			yield elem_2.r_data.eq(0xaa)

			yield bus.addr.eq(elem_1_addr)
			yield bus.r_stb.eq(1)
			yield
			yield bus.addr.eq(elem_2_addr)
			yield
			self.assertEqual((yield bus.r_data), 0x55)
			yield
			self.assertEqual((yield bus.r_data), 0xaa)

		m = Module()
		m.submodules += self.dut, mux_1, mux_2
		sim = Simulator(m)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()



# SPDX-License-Identifier: BSD-2-Clause


from collections import OrderedDict
from enum        import Enum
from typing      import Generator, Optional

from ...         import Elaboratable, Module, Record, Signal

__all__ = (
	'EventMap',
	'Monitor',
	'Source',
)


class Source(Record):
	class Trigger(Enum):
		''' Event trigger mode. '''
		LEVEL = 'level'
		RISE  = 'rise'
		FALL  = 'fall'

	'''
	Event source interface.

	Parameters
	----------
	trigger : :class:`Trigger`
		Trigger mode. An event can be edge- or level-triggered by the input line.
	name: str
		Name of the underlying record.

	Attributes
	----------
	i : Signal()
		Input line. Sampled in order to detect an event.
	trg : Signal()
		Event trigger. Asserted when an event occurs, according to the trigger mode.

	'''

	def __init__(
		self, *, trigger: Trigger = 'level', name: Optional[str] = None, src_loc_at: int = 0
	) -> None:
		choices = ("level", 'rise', 'fall')
		if not isinstance(trigger, Source.Trigger) and trigger not in choices:
			raise ValueError(f'Invalid trigger mode {trigger!r}; must be one of {", ".join(choices)}')
		self.trigger = Source.Trigger(trigger)
		self._map    = None

		super().__init__([
			('i',   1),
			('trg', 1),
		], name = name, src_loc_at = 1 + src_loc_at)

	@property
	def event_map(self) -> 'EventMap':
		'''
		Event map.

		Return value
		------------
		A :class:`EventMap` describing subordinate event sources.

		Exceptions
		----------
		Raises :exn:`NotImplementedError` if the source does not have an event map.

		'''

		if self._map is None:
			raise NotImplementedError(f'Event source {self!r} does not have an event map')
		return self._map

	@event_map.setter
	def event_map(self, event_map: 'EventMap') -> None:
		if not isinstance(event_map, EventMap):
			raise TypeError(f'Event map must be an instance of EventMap, not {event_map!r}')
		event_map.freeze()
		self._map = event_map

	# FIXME: get rid of this
	__hash__ = object.__hash__


class EventMap:
	'''
	Event map.

	An event map is a description of a set of events. It is built by adding event sources
	and can be queried later to determine their index. Event indexing is done implicitly by
	increment, starting at 0.

	'''
	def __init__(self) -> None:
		self._sources = OrderedDict()
		self._frozen  = False

	@property
	def size(self) -> int:
		'''
		Size of the event map.

		Return value
		------------
		The number of event sources in the map.

		'''
		return len(self._sources)

	def freeze(self) -> None:
		'''
		Freeze the event map.

		Once the event map is frozen, sources cannot be added anymore.

		'''

		self._frozen = True

	def add(self, src: Source) -> None:
		'''
		Add an event source.

		Arguments
		---------
		src : :class:`Source`
			Event source.

		Exceptions
		----------
		Raises :exn:`ValueError` if the event map is frozen.

		'''

		if self._frozen:
			raise ValueError('Event map has been frozen. Cannot add source.')
		if not isinstance(src, Source):
			raise TypeError(f'Event source must be an instance of event.Source, not {src!r}')
		if src not in self._sources:
			self._sources[src] = self.size

	def index(self, src: Source) -> int:
		'''
		Get the index corresponding to an event source.

		Arguments
		---------
		src : :class:`Source`
			Event source.

		Return value
		------------
		The index of the source.

		Exceptions
		----------
		Raises :exn:`KeyError` if the source is not found.

		'''

		if not isinstance(src, Source):
			raise TypeError(f'Event source must be an instance of event.Source, not {src!r}')
		return self._sources[src]

	def sources(self) -> Generator[tuple[Source, int], None, None]:
		'''
		Iterate event sources.

		Yield values
		------------
		A tuple ``src, index`` corresponding to an event source and its index.

		'''

		for src, index in self._sources.items():
			yield (src, index)


class Monitor(Elaboratable):
	'''
	Event monitor.

	A monitor for subordinate event sources.

	Parameters
	----------
	event_map : :class:`EventMap`
		Event map.
	trigger : :class:`Source.Trigger`
		Trigger mode. See :class:`Source`.

	Attributes
	----------
	src : :class:`Source`
		Event source. Its input is asserted when a subordinate event is enabled and pending.
	enable : Signal(event_map.size), bit mask, in
		Enabled events.
	pending : Signal(event_map.size), bit mask, out
		Pending events.
	clear : Signal(event_map.size), bit mask, in
		Clear selected pending events.

	'''

	def __init__(self, event_map: EventMap, *, trigger: Source.Trigger = 'level') -> None:
		self.src = Source(trigger = trigger)
		self.src.event_map = event_map

		self.enable  = Signal(event_map.size)
		self.pending = Signal(event_map.size)
		self.clear   = Signal(event_map.size)

	def elaborate(self, platform) -> Module:
		m = Module()

		for sub, index in self.src.event_map.sources():
			if sub.trigger != Source.Trigger.LEVEL:
				sub_i_r = Signal.like(sub.i, name_suffix = '_r')
				m.d.sync += sub_i_r.eq(sub.i)

			if sub.trigger == Source.Trigger.LEVEL:
				m.d.comb += sub.trg.eq(sub.i)
			elif sub.trigger == Source.Trigger.RISE:
				m.d.comb += sub.trg.eq(~sub_i_r & sub.i)
			elif sub.trigger == Source.Trigger.FALL:
				m.d.comb += sub.trg.eq( sub_i_r & ~sub.i)
			else:
				assert False # :nocov:

			with m.If(sub.trg):
				m.d.sync += self.pending[index].eq(1)
			with m.Elif(self.clear[index]):
				m.d.sync += self.pending[index].eq(0)

		m.d.comb += self.src.i.eq((self.enable & self.pending).any())

		return m



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, Record
from torii.back import verilog


class FlatGPIO(Elaboratable):
	def __init__(self, pins, bus):
		self.pins = pins
		self.bus  = bus

	def elaborate(self, platform) -> Module:
		bus  = self.bus

		m = Module()
		m.d.comb += bus.r_data.eq(self.pins.word_select(bus.addr, len(bus.r_data)))
		with m.If(bus.we):
			m.d.sync += self.pins.word_select(bus.addr, len(bus.w_data)).eq(bus.w_data)
		return m


if __name__ == '__main__':
	bus = Record([
		('addr',   3),
		('r_data', 2),
		('w_data', 2),
		('we',     1),
	])
	pins = Signal(8)
	gpio = FlatGPIO(pins, bus)

	print(verilog.convert(gpio, ports = [pins, bus.addr, bus.r_data, bus.w_data, bus.we]))



# SPDX-License-Identifier: BSD-2-Clause

from ....hdl import Elaboratable, Instance, Module, Signal

__all__ = (
	'ice40Warmboot',
)

class ice40Warmboot(Elaboratable):
	'''
	Lattice ice40 Warmboot block

	Attributes
	----------

	boot_sel : Signal(2)
		boot slot selection

	boot_strb: Signal
		Strobe to trigger warmboot

	'''

	def __init__(self) -> None:
		self.boot_sel  = Signal(2)
		self.boot_strb = Signal()

	def elaborate(self, _) -> Module:
		m = Module()

		m.submodules.ice40_warmboot = Instance(
			'SB_WARMBOOT',
			i_BOOT = self.boot_strb,
			i_S0   = self.boot_sel[0],
			i_S1   = self.boot_sel[1]
		)

		return m



# SPDX-License-Identifier: BSD-2-Clause

from abc             import ABCMeta, abstractmethod
from collections     import OrderedDict
from collections.abc import Iterable

from ..util          import flatten, tracer
from .ast            import (
	AnyConst, AnySeq, ArrayProxy, Assert, Assign, Assume, Cat,
	ClockSignal, Const, Cover, Initial, Mux, Operator, Part,
	Property, ResetSignal, Sample, Signal, SignalDict,
	SignalSet, Slice, Statement, Switch, Value, ValueDict,
	_StatementList
)
from .cd             import DomainError
from .ir             import Elaboratable, Fragment, Instance

__all__ = (
	'DomainCollector',
	'DomainLowerer',
	'DomainRenamer',
	'EnableInserter',
	'FragmentTransformer',
	'LHSGroupAnalyzer',
	'LHSGroupFilter',
	'ResetInserter',
	'SampleDomainInjector',
	'SampleLowerer',
	'StatementTransformer',
	'StatementVisitor',
	'SwitchCleaner',
	'TransformedElaboratable',
	'ValueTransformer',
	'ValueVisitor',
)


class ValueVisitor(metaclass = ABCMeta):
	@abstractmethod
	def on_Const(self, value):
		pass # :nocov:

	@abstractmethod
	def on_AnyConst(self, value):
		pass # :nocov:

	@abstractmethod
	def on_AnySeq(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Signal(self, value):
		pass # :nocov:

	@abstractmethod
	def on_ClockSignal(self, value):
		pass # :nocov:

	@abstractmethod
	def on_ResetSignal(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Operator(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Slice(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Part(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Cat(self, value):
		pass # :nocov:

	@abstractmethod
	def on_ArrayProxy(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Sample(self, value):
		pass # :nocov:

	@abstractmethod
	def on_Initial(self, value):
		pass # :nocov:

	def on_unknown_value(self, value):
		raise TypeError(f'Cannot transform value {value!r}') # :nocov:

	def replace_value_src_loc(self, value, new_value):
		return True

	def on_value(self, value):
		if type(value) is Const:
			new_value = self.on_Const(value)
		elif type(value) is AnyConst:
			new_value = self.on_AnyConst(value)
		elif type(value) is AnySeq:
			new_value = self.on_AnySeq(value)
		elif type(value) is Signal:
			new_value = self.on_Signal(value)
		elif type(value) is ClockSignal:
			new_value = self.on_ClockSignal(value)
		elif type(value) is ResetSignal:
			new_value = self.on_ResetSignal(value)
		elif type(value) is Operator:
			new_value = self.on_Operator(value)
		elif type(value) is Slice:
			new_value = self.on_Slice(value)
		elif type(value) is Part:
			new_value = self.on_Part(value)
		elif type(value) is Cat:
			new_value = self.on_Cat(value)
		elif type(value) is ArrayProxy:
			new_value = self.on_ArrayProxy(value)
		elif type(value) is Sample:
			new_value = self.on_Sample(value)
		elif type(value) is Initial:
			new_value = self.on_Initial(value)
		else:
			new_value = self.on_unknown_value(value)
		if isinstance(new_value, Value) and self.replace_value_src_loc(value, new_value):
			new_value.src_loc = value.src_loc
		return new_value

	def __call__(self, value):
		return self.on_value(value)


class ValueTransformer(ValueVisitor):
	def on_Const(self, value):
		return value

	def on_AnyConst(self, value):
		return value

	def on_AnySeq(self, value):
		return value

	def on_Signal(self, value):
		return value

	def on_ClockSignal(self, value):
		return value

	def on_ResetSignal(self, value):
		return value

	def on_Operator(self, value):
		return Operator(value.operator, [ self.on_value(o) for o in value.operands ])

	def on_Slice(self, value):
		return Slice(self.on_value(value.value), value.start, value.stop)

	def on_Part(self, value):
		return Part(self.on_value(value.value), self.on_value(value.offset),
					value.width, value.stride)

	def on_Cat(self, value):
		return Cat(self.on_value(o) for o in value.parts)

	def on_ArrayProxy(self, value):
		return ArrayProxy(
			[self.on_value(elem) for elem in value._iter_as_values()],
			self.on_value(value.index)
		)

	def on_Sample(self, value):
		return Sample(self.on_value(value.value), value.clocks, value.domain)

	def on_Initial(self, value):
		return value


class StatementVisitor(metaclass = ABCMeta):
	@abstractmethod
	def on_Assign(self, stmt):
		pass # :nocov:

	@abstractmethod
	def on_Assert(self, stmt):
		pass # :nocov:

	@abstractmethod
	def on_Assume(self, stmt):
		pass # :nocov:

	@abstractmethod
	def on_Cover(self, stmt):
		pass # :nocov:

	@abstractmethod
	def on_Switch(self, stmt):
		pass # :nocov:

	@abstractmethod
	def on_statements(self, stmts):
		pass # :nocov:

	def on_unknown_statement(self, stmt):
		raise TypeError(f'Cannot transform statement {stmt!r}') # :nocov:

	def replace_statement_src_loc(self, stmt, new_stmt):
		return True

	def on_statement(self, stmt):
		if type(stmt) is Assign:
			new_stmt = self.on_Assign(stmt)
		elif type(stmt) is Assert:
			new_stmt = self.on_Assert(stmt)
		elif type(stmt) is Assume:
			new_stmt = self.on_Assume(stmt)
		elif type(stmt) is Cover:
			new_stmt = self.on_Cover(stmt)
		elif isinstance(stmt, Switch):
			# Uses `isinstance()` and not `type() is` because amaranth.compat requires it.
			new_stmt = self.on_Switch(stmt)
		elif isinstance(stmt, Iterable):
			new_stmt = self.on_statements(stmt)
		else:
			new_stmt = self.on_unknown_statement(stmt)
		if isinstance(new_stmt, Statement) and self.replace_statement_src_loc(stmt, new_stmt):
			new_stmt.src_loc = stmt.src_loc
			if isinstance(new_stmt, Switch) and isinstance(stmt, Switch):
				new_stmt.case_src_locs = stmt.case_src_locs
		if isinstance(new_stmt, Property):
			new_stmt._MustUse__used = True
		return new_stmt

	def __call__(self, stmt):
		return self.on_statement(stmt)


class StatementTransformer(StatementVisitor):
	def on_value(self, value):
		return value

	def on_Assign(self, stmt):
		return Assign(self.on_value(stmt.lhs), self.on_value(stmt.rhs))

	def on_Assert(self, stmt):
		return Assert(
			self.on_value(stmt.test), _check = stmt._check,
			_en = stmt._en, name = stmt.name
		)

	def on_Assume(self, stmt):
		return Assume(
			self.on_value(stmt.test), _check = stmt._check,
			_en = stmt._en, name = stmt.name
		)

	def on_Cover(self, stmt):
		return Cover(
			self.on_value(stmt.test), _check = stmt._check,
			_en = stmt._en, name = stmt.name
		)

	def on_Switch(self, stmt):
		cases = OrderedDict((k, self.on_statement(s)) for k, s in stmt.cases.items())
		return Switch(self.on_value(stmt.test), cases)

	def on_statements(self, stmts):
		return _StatementList(flatten(self.on_statement(stmt) for stmt in stmts))


class FragmentTransformer:
	def map_subfragments(self, fragment, new_fragment):
		for subfragment, name in fragment.subfragments:
			new_fragment.add_subfragment(self(subfragment), name)

	def map_ports(self, fragment, new_fragment):
		for port, dir in fragment.ports.items():
			new_fragment.add_ports(port, dir = dir)

	def map_named_ports(self, fragment, new_fragment):
		if hasattr(self, 'on_value'):
			for name, (value, dir) in fragment.named_ports.items():
				new_fragment.named_ports[name] = self.on_value(value), dir
		else:
			new_fragment.named_ports = OrderedDict(fragment.named_ports.items())

	def map_domains(self, fragment, new_fragment):
		for domain in fragment.iter_domains():
			new_fragment.add_domains(fragment.domains[domain])

	def map_statements(self, fragment, new_fragment):
		if hasattr(self, 'on_statement'):
			new_fragment.add_statements(map(self.on_statement, fragment.statements))
		else:
			new_fragment.add_statements(fragment.statements)

	def map_drivers(self, fragment, new_fragment):
		for domain, signal in fragment.iter_drivers():
			new_fragment.add_driver(signal, domain)

	def on_fragment(self, fragment):
		if isinstance(fragment, Instance):
			new_fragment = Instance(fragment.type)
			new_fragment.parameters = OrderedDict(fragment.parameters)
			self.map_named_ports(fragment, new_fragment)
		else:
			new_fragment = Fragment()
			new_fragment.flatten = fragment.flatten
		new_fragment.attrs = OrderedDict(fragment.attrs)
		self.map_ports(fragment, new_fragment)
		self.map_subfragments(fragment, new_fragment)
		self.map_domains(fragment, new_fragment)
		self.map_statements(fragment, new_fragment)
		self.map_drivers(fragment, new_fragment)
		return new_fragment

	def __call__(self, value, *, src_loc_at = 0):
		if isinstance(value, Fragment):
			return self.on_fragment(value)
		elif isinstance(value, TransformedElaboratable):
			value._transforms_.append(self)
			return value
		elif hasattr(value, 'elaborate'):
			value = TransformedElaboratable(value, src_loc_at = 1 + src_loc_at)
			value._transforms_.append(self)
			return value
		else:
			raise AttributeError(f'Object {value!r} cannot be elaborated')


class TransformedElaboratable(Elaboratable):
	def __init__(self, elaboratable, *, src_loc_at = 0):
		assert hasattr(elaboratable, 'elaborate')

		# Fields prefixed and suffixed with underscore to avoid as many conflicts with the inner
		# object as possible, since we're forwarding attribute requests to it.
		self._elaboratable_ = elaboratable
		self._transforms_   = []

	def __getattr__(self, attr):
		return getattr(self._elaboratable_, attr)

	def elaborate(self, platform):
		fragment = Fragment.get(self._elaboratable_, platform)
		for transform in self._transforms_:
			fragment = transform(fragment)
		return fragment


class DomainCollector(ValueVisitor, StatementVisitor):
	def __init__(self):
		self.used_domains = set()
		self.defined_domains = set()
		self._local_domains = set()

	def _add_used_domain(self, domain_name):
		if domain_name is None:
			return
		if domain_name in self._local_domains:
			return
		self.used_domains.add(domain_name)

	def on_ignore(self, value):
		pass

	on_Const = on_ignore
	on_AnyConst = on_ignore
	on_AnySeq = on_ignore
	on_Signal = on_ignore

	def on_ClockSignal(self, value):
		self._add_used_domain(value.domain)

	def on_ResetSignal(self, value):
		self._add_used_domain(value.domain)

	def on_Operator(self, value):
		for o in value.operands:
			self.on_value(o)

	def on_Slice(self, value):
		self.on_value(value.value)

	def on_Part(self, value):
		self.on_value(value.value)
		self.on_value(value.offset)

	def on_Cat(self, value):
		for o in value.parts:
			self.on_value(o)

	def on_ArrayProxy(self, value):
		for elem in value._iter_as_values():
			self.on_value(elem)
		self.on_value(value.index)

	def on_Sample(self, value):
		self.on_value(value.value)

	def on_Initial(self, value):
		pass

	def on_Assign(self, stmt):
		self.on_value(stmt.lhs)
		self.on_value(stmt.rhs)

	def on_property(self, stmt):
		self.on_value(stmt.test)

	on_Assert = on_property
	on_Assume = on_property
	on_Cover  = on_property

	def on_Switch(self, stmt):
		self.on_value(stmt.test)
		for stmts in stmt.cases.values():
			self.on_statement(stmts)

	def on_statements(self, stmts):
		for stmt in stmts:
			self.on_statement(stmt)

	def on_fragment(self, fragment):
		if isinstance(fragment, Instance):
			for name, (value, dir) in fragment.named_ports.items():
				self.on_value(value)

		old_local_domains, self._local_domains = self._local_domains, set(self._local_domains)
		for domain_name, domain in fragment.domains.items():
			if domain.local:
				self._local_domains.add(domain_name)
			else:
				self.defined_domains.add(domain_name)

		self.on_statements(fragment.statements)
		for domain_name in fragment.drivers:
			self._add_used_domain(domain_name)
		for subfragment, name in fragment.subfragments:
			self.on_fragment(subfragment)

		self._local_domains = old_local_domains

	def __call__(self, fragment):
		self.on_fragment(fragment)


class DomainRenamer(FragmentTransformer, ValueTransformer, StatementTransformer):
	def __init__(self, domain_map):
		if isinstance(domain_map, str):
			domain_map = { 'sync': domain_map }
		for src, dst in domain_map.items():
			if src == 'comb':
				raise ValueError(f'Domain \'{src}\' may not be renamed')
			if dst == 'comb':
				raise ValueError(f'Domain \'{src}\' may not be renamed to \'{dst}\'')
		self.domain_map = OrderedDict(domain_map)

	def on_ClockSignal(self, value):
		if value.domain in self.domain_map:
			return ClockSignal(self.domain_map[value.domain])
		return value

	def on_ResetSignal(self, value):
		if value.domain in self.domain_map:
			return ResetSignal(
				self.domain_map[value.domain],
				allow_reset_less = value.allow_reset_less
			)
		return value

	def map_domains(self, fragment, new_fragment):
		for domain in fragment.iter_domains():
			cd = fragment.domains[domain]
			if domain in self.domain_map:
				if cd.name == domain:
					# Rename the actual ClockDomain object.
					cd.rename(self.domain_map[domain])
				else:
					assert cd.name == self.domain_map[domain]
			new_fragment.add_domains(cd)

	def map_drivers(self, fragment, new_fragment):
		for domain, signals in fragment.drivers.items():
			if domain in self.domain_map:
				domain = self.domain_map[domain]
			for signal in signals:
				new_fragment.add_driver(self.on_value(signal), domain)


class DomainLowerer(FragmentTransformer, ValueTransformer, StatementTransformer):
	def __init__(self, domains = None):
		self.domains = domains

	def _resolve(self, domain, context):
		if domain not in self.domains:
			raise DomainError(f'Signal {context!r} refers to nonexistent domain \'{domain}\'')
		return self.domains[domain]

	def map_drivers(self, fragment, new_fragment):
		for domain, signal in fragment.iter_drivers():
			new_fragment.add_driver(self.on_value(signal), domain)

	def replace_value_src_loc(self, value, new_value):
		return not isinstance(value, (ClockSignal, ResetSignal))

	def on_ClockSignal(self, value):
		domain = self._resolve(value.domain, value)
		return domain.clk

	def on_ResetSignal(self, value):
		domain = self._resolve(value.domain, value)
		if domain.rst is None:
			if value.allow_reset_less:
				return Const(0)
			else:
				raise DomainError(f'Signal {value!r} refers to reset of reset-less domain \'{value.domain}\'')
		return domain.rst

	def _insert_resets(self, fragment):
		for domain_name, signals in fragment.drivers.items():
			if domain_name is None:
				continue
			domain = fragment.domains[domain_name]
			if domain.rst is None:
				continue
			stmts = [
				signal.eq(Const(signal.reset, signal.width))
				for signal in signals if not signal.reset_less
			]
			fragment.add_statements(Switch(domain.rst, {1: stmts}))

	def on_fragment(self, fragment):
		self.domains = fragment.domains
		new_fragment = super().on_fragment(fragment)
		self._insert_resets(new_fragment)
		return new_fragment


class SampleDomainInjector(ValueTransformer, StatementTransformer):
	def __init__(self, domain):
		self.domain = domain

	def on_Sample(self, value):
		if value.domain is not None:
			return value
		return Sample(value.value, value.clocks, self.domain)

	def __call__(self, stmts):
		return self.on_statement(stmts)


class SampleLowerer(FragmentTransformer, ValueTransformer, StatementTransformer):
	def __init__(self):
		self.initial = None
		self.sample_cache = None
		self.sample_stmts = None

	def _name_reset(self, value):
		if isinstance(value, Const):
			return f'c${value.value}', value.value
		elif isinstance(value, Signal):
			return f's${value.name}', value.reset
		elif isinstance(value, ClockSignal):
			return 'clk', 0
		elif isinstance(value, ResetSignal):
			return 'rst', 1
		elif isinstance(value, Initial):
			return 'init', 0 # Past(Initial()) produces 0, 1, 0, 0, ...
		else:
			raise NotImplementedError # :nocov:

	def on_Sample(self, value):
		if value in self.sample_cache:
			return self.sample_cache[value]

		sampled_value = self.on_value(value.value)
		if value.clocks == 0:
			sample = sampled_value
		else:
			assert value.domain is not None
			sampled_name, sampled_reset = self._name_reset(value.value)
			name = f'$sample${sampled_name}${value.domain}${value.clocks}'
			sample = Signal.like(value.value, name = name, reset_less = True, reset = sampled_reset)
			sample.attrs['torii.sample_reg'] = True

			prev_sample = self.on_Sample(Sample(sampled_value, value.clocks - 1, value.domain))
			if value.domain not in self.sample_stmts:
				self.sample_stmts[value.domain] = []
			self.sample_stmts[value.domain].append(sample.eq(prev_sample))

		self.sample_cache[value] = sample
		return sample

	def on_Initial(self, value):
		if self.initial is None:
			self.initial = Signal(name = 'init')
		return self.initial

	def map_statements(self, fragment, new_fragment):
		self.initial = None
		self.sample_cache = ValueDict()
		self.sample_stmts = OrderedDict()
		new_fragment.add_statements(map(self.on_statement, fragment.statements))
		for domain, stmts in self.sample_stmts.items():
			new_fragment.add_statements(stmts)
			for stmt in stmts:
				new_fragment.add_driver(stmt.lhs, domain)
		if self.initial is not None:
			new_fragment.add_subfragment(Instance('$initstate', o_Y = self.initial))


class SwitchCleaner(StatementVisitor):
	def on_ignore(self, stmt):
		return stmt

	on_Assign = on_ignore
	on_Assert = on_ignore
	on_Assume = on_ignore
	on_Cover  = on_ignore

	def on_Switch(self, stmt):
		cases = OrderedDict((k, self.on_statement(s)) for k, s in stmt.cases.items())
		if any(len(s) for s in cases.values()):
			return Switch(stmt.test, cases)

	def on_statements(self, stmts):
		stmts = flatten(self.on_statement(stmt) for stmt in stmts)
		return _StatementList(stmt for stmt in stmts if stmt is not None)


class LHSGroupAnalyzer(StatementVisitor):
	def __init__(self):
		self.signals = SignalDict()
		self.unions  = OrderedDict()

	def find(self, signal):
		if signal not in self.signals:
			self.signals[signal] = len(self.signals)
		group = self.signals[signal]
		while group in self.unions:
			group = self.unions[group]
		self.signals[signal] = group
		return group

	def unify(self, root, *leaves):
		root_group = self.find(root)
		for leaf in leaves:
			leaf_group = self.find(leaf)
			if root_group == leaf_group:
				continue
			self.unions[leaf_group] = root_group

	def groups(self):
		groups = OrderedDict()
		for signal in self.signals:
			group = self.find(signal)
			if group not in groups:
				groups[group] = SignalSet()
			groups[group].add(signal)
		return groups

	def on_Assign(self, stmt):
		lhs_signals = stmt._lhs_signals()
		if lhs_signals:
			self.unify(*stmt._lhs_signals())

	def on_property(self, stmt):
		lhs_signals = stmt._lhs_signals()
		if lhs_signals:
			self.unify(*stmt._lhs_signals())

	on_Assert = on_property
	on_Assume = on_property
	on_Cover  = on_property

	def on_Switch(self, stmt):
		for case_stmts in stmt.cases.values():
			self.on_statements(case_stmts)

	def on_statements(self, stmts):
		for stmt in stmts:
			self.on_statement(stmt)

	def __call__(self, stmts):
		self.on_statements(stmts)
		return self.groups()


class LHSGroupFilter(SwitchCleaner):
	def __init__(self, signals):
		self.signals = signals

	def on_Assign(self, stmt):
		# The invariant provided by LHSGroupAnalyzer is that all signals that ever appear together
		# on LHS are a part of the same group, so it is sufficient to check any of them.
		lhs_signals = stmt.lhs._lhs_signals()
		if lhs_signals:
			any_lhs_signal = next(iter(lhs_signals))
			if any_lhs_signal in self.signals:
				return stmt

	def on_property(self, stmt):
		any_lhs_signal = next(iter(stmt._lhs_signals()))
		if any_lhs_signal in self.signals:
			return stmt

	on_Assert = on_property
	on_Assume = on_property
	on_Cover  = on_property


class _ControlInserter(FragmentTransformer):
	def __init__(self, controls):
		self.src_loc = None
		if isinstance(controls, Value):
			controls = { 'sync': controls }
		self.controls = OrderedDict(controls)

	def on_fragment(self, fragment):
		new_fragment = super().on_fragment(fragment)
		for domain, signals in fragment.drivers.items():
			if domain is None or domain not in self.controls:
				continue
			self._insert_control(new_fragment, domain, signals)
		return new_fragment

	def _insert_control(self, fragment, domain, signals):
		raise NotImplementedError # :nocov:

	def __call__(self, value, *, src_loc_at = 0):
		self.src_loc = tracer.get_src_loc(src_loc_at = src_loc_at)
		return super().__call__(value, src_loc_at = 1 + src_loc_at)


class ResetInserter(_ControlInserter):
	def _insert_control(self, fragment, domain, signals):
		stmts = [ s.eq(Const(s.reset, s.width)) for s in signals if not s.reset_less ]
		fragment.add_statements(Switch(self.controls[domain], { 1: stmts }, src_loc = self.src_loc))


class EnableInserter(_ControlInserter):
	def _insert_control(self, fragment, domain, signals):
		stmts = [s.eq(s) for s in signals]
		fragment.add_statements(Switch(self.controls[domain], { 0: stmts }, src_loc = self.src_loc))

	def on_fragment(self, fragment):
		new_fragment = super().on_fragment(fragment)
		if isinstance(new_fragment, Instance) and new_fragment.type == '$mem_v2':
			for kind in ('RD', 'WR'):
				clk_parts = new_fragment.named_ports[f'{kind}_CLK'][0].parts
				en_parts  = new_fragment.named_ports[f'{kind}_EN'][0].parts
				new_en = []
				for clk, en in zip(clk_parts, en_parts):
					if isinstance(clk, ClockSignal) and clk.domain in self.controls:
						en = Mux(self.controls[clk.domain], en, Const(0, len(en)))
					new_en.append(en)
				new_fragment.named_ports[f'{kind}_EN'] = (Cat(new_en), 'i')
		return new_fragment



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from collections   import OrderedDict

from torii.hdl.ast import *
from torii.hdl.cd  import *
from torii.hdl.ir  import *
from torii.hdl.mem import *

from ..utils       import ToriiTestSuiteCase


class ElaboratesToNone(Elaboratable):
	def elaborate(self, platform):
		return


class ElaboratesToSelf(Elaboratable):
	def elaborate(self, platform):
		return self


class FragmentGetTestCase(ToriiTestSuiteCase):
	def test_get_wrong_none(self):
		with self.assertRaisesRegex(
			AttributeError,
			r'^Object None cannot be elaborated$'
		):
			Fragment.get(None, platform = None)

		with self.assertWarnsRegex(
			UserWarning,
			r'^\.elaborate\(\) returned None; missing return statement\?$'
		):
			with self.assertRaisesRegex(
				AttributeError,
				r'^Object None cannot be elaborated$'
			):
				Fragment.get(ElaboratesToNone(), platform = None)

	def test_get_wrong_self(self):
		with self.assertRaisesRegex(
			RecursionError,
			r'^Object <.+?ElaboratesToSelf.+?> elaborates to itself$'
		):
			Fragment.get(ElaboratesToSelf(), platform = None)


class FragmentGeneratedTestCase(ToriiTestSuiteCase):
	def test_find_subfragment(self):
		f1 = Fragment()
		f2 = Fragment()
		f1.add_subfragment(f2, 'f2')

		self.assertEqual(f1.find_subfragment(0), f2)
		self.assertEqual(f1.find_subfragment('f2'), f2)

	def test_find_subfragment_wrong(self):
		f1 = Fragment()
		f2 = Fragment()
		f1.add_subfragment(f2, 'f2')

		with self.assertRaisesRegex(
			NameError,
			r'^No subfragment at index #1$'
		):
			f1.find_subfragment(1)
		with self.assertRaisesRegex(
			NameError,
			r'^No subfragment with name \'fx\'$'
		):
			f1.find_subfragment('fx')

	def test_find_generated(self):
		f1 = Fragment()
		f2 = Fragment()
		f2.generated['sig'] = sig = Signal()
		f1.add_subfragment(f2, 'f2')

		self.assertEqual(
			SignalKey(f1.find_generated('f2', 'sig')),
			SignalKey(sig)
		)


class FragmentDriversTestCase(ToriiTestSuiteCase):
	def test_empty(self):
		f = Fragment()
		self.assertEqual(list(f.iter_comb()), [])
		self.assertEqual(list(f.iter_sync()), [])


class FragmentPortsTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.s1 = Signal()
		self.s2 = Signal()
		self.s3 = Signal()
		self.c1 = Signal()
		self.c2 = Signal()
		self.c3 = Signal()

	def test_empty(self):
		f = Fragment()
		self.assertEqual(list(f.iter_ports()), [])

		f._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([]))

	def test_iter_signals(self):
		f = Fragment()
		f.add_ports(self.s1, self.s2, dir = 'io')
		self.assertEqual(SignalSet((self.s1, self.s2)), f.iter_signals())

	def test_self_contained(self):
		f = Fragment()
		f.add_statements(
			self.c1.eq(self.s1),
			self.s1.eq(self.c1)
		)

		f._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([]))

	def test_infer_input(self):
		f = Fragment()
		f.add_statements(
			self.c1.eq(self.s1)
		)

		f._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([
			(self.s1, 'i')
		]))

	def test_request_output(self):
		f = Fragment()
		f.add_statements(
			self.c1.eq(self.s1)
		)

		f._propagate_ports(ports = (self.c1,), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([
			(self.s1, 'i'),
			(self.c1, 'o')
		]))

	def test_input_in_subfragment(self):
		f1 = Fragment()
		f1.add_statements(
			self.c1.eq(self.s1)
		)
		f2 = Fragment()
		f2.add_statements(
			self.s1.eq(0)
		)
		f1.add_subfragment(f2)
		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict())
		self.assertEqual(f2.ports, SignalDict([
			(self.s1, 'o'),
		]))

	def test_input_only_in_subfragment(self):
		f1 = Fragment()
		f2 = Fragment()
		f2.add_statements(
			self.c1.eq(self.s1)
		)
		f1.add_subfragment(f2)
		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict([
			(self.s1, 'i'),
		]))
		self.assertEqual(f2.ports, SignalDict([
			(self.s1, 'i'),
		]))

	def test_output_from_subfragment(self):
		f1 = Fragment()
		f1.add_statements(
			self.c1.eq(0)
		)
		f2 = Fragment()
		f2.add_statements(
			self.c2.eq(1)
		)
		f1.add_subfragment(f2)

		f1._propagate_ports(ports = (self.c2,), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict([
			(self.c2, 'o'),
		]))
		self.assertEqual(f2.ports, SignalDict([
			(self.c2, 'o'),
		]))

	def test_output_from_subfragment_2(self):
		f1 = Fragment()
		f1.add_statements(
			self.c1.eq(self.s1)
		)
		f2 = Fragment()
		f2.add_statements(
			self.c2.eq(self.s1)
		)
		f1.add_subfragment(f2)
		f3 = Fragment()
		f3.add_statements(
			self.s1.eq(0)
		)
		f2.add_subfragment(f3)

		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f2.ports, SignalDict([
			(self.s1, 'o'),
		]))

	def test_input_output_sibling(self):
		f1 = Fragment()
		f2 = Fragment()
		f2.add_statements(
			self.c1.eq(self.c2)
		)
		f1.add_subfragment(f2)
		f3 = Fragment()
		f3.add_statements(
			self.c2.eq(0)
		)
		f3.add_driver(self.c2)
		f1.add_subfragment(f3)

		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict())

	def test_output_input_sibling(self):
		f1 = Fragment()
		f2 = Fragment()
		f2.add_statements(
			self.c2.eq(0)
		)
		f2.add_driver(self.c2)
		f1.add_subfragment(f2)
		f3 = Fragment()
		f3.add_statements(
			self.c1.eq(self.c2)
		)
		f1.add_subfragment(f3)

		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict())

	def test_input_cd(self):
		sync = ClockDomain()
		f = Fragment()
		f.add_statements(
			self.c1.eq(self.s1)
		)
		f.add_domains(sync)
		f.add_driver(self.c1, 'sync')

		f._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([
			(self.s1,  'i'),
			(sync.clk, 'i'),
			(sync.rst, 'i'),
		]))

	def test_input_cd_reset_less(self):
		sync = ClockDomain(reset_less = True)
		f = Fragment()
		f.add_statements(
			self.c1.eq(self.s1)
		)
		f.add_domains(sync)
		f.add_driver(self.c1, 'sync')

		f._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f.ports, SignalDict([
			(self.s1,  'i'),
			(sync.clk, 'i'),
		]))

	def test_inout(self):
		s = Signal()
		f1 = Fragment()
		f2 = Instance('foo', io_x = s)
		f1.add_subfragment(f2)

		f1._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict([
			(s, 'io')
		]))

	def test_in_out_same_signal(self):
		s = Signal()

		f1 = Instance('foo', i_x = s, o_y = s)
		f2 = Fragment()
		f2.add_subfragment(f1)

		f2._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f1.ports, SignalDict([
			(s, 'o')
		]))

		f3 = Instance('foo', o_y = s, i_x = s)
		f4 = Fragment()
		f4.add_subfragment(f3)

		f4._propagate_ports(ports = (), all_undef_as_ports = True)
		self.assertEqual(f3.ports, SignalDict([
			(s, 'o')
		]))

	def test_clk_rst(self):
		sync = ClockDomain()
		f = Fragment()
		f.add_domains(sync)

		f = f.prepare(ports = (ClockSignal('sync'), ResetSignal('sync')))
		self.assertEqual(f.ports, SignalDict([
			(sync.clk, 'i'),
			(sync.rst, 'i'),
		]))

	def test_port_wrong(self):
		f = Fragment()
		with self.assertRaisesRegex(
			TypeError,
			r'^Only signals may be added as ports, not \(const 1\'d1\)$'
		):
			f.prepare(ports = (Const(1),))

	def test_port_not_iterable(self):
		f = Fragment()
		with self.assertRaisesRegex(
			TypeError,
			r'^`ports` must be either a list or a tuple, not 1$'
		):
			f.prepare(ports = 1)
		with self.assertRaisesRegex(
			TypeError, (
				r'^`ports` must be either a list or a tuple, not \(const 1\'d1\)'
				r' \(did you mean `ports = \(<signal>,\)`, rather than `ports = <signal>`\?\)$'
			)
		):
			f.prepare(ports = Const(1))

class FragmentDomainsTestCase(ToriiTestSuiteCase):
	def test_iter_signals(self):
		cd1 = ClockDomain()
		cd2 = ClockDomain(reset_less = True)
		s1 = Signal()
		s2 = Signal()

		f = Fragment()
		f.add_domains(cd1, cd2)
		f.add_driver(s1, 'cd1')
		self.assertEqual(SignalSet((cd1.clk, cd1.rst, s1)), f.iter_signals())
		f.add_driver(s2, 'cd2')
		self.assertEqual(SignalSet((cd1.clk, cd1.rst, cd2.clk, s1, s2)), f.iter_signals())

	def test_propagate_up(self):
		cd = ClockDomain()

		f1 = Fragment()
		f2 = Fragment()
		f1.add_subfragment(f2)
		f2.add_domains(cd)

		f1._propagate_domains_up()
		self.assertEqual(f1.domains, {'cd': cd})

	def test_propagate_up_local(self):
		cd = ClockDomain(local = True)

		f1 = Fragment()
		f2 = Fragment()
		f1.add_subfragment(f2)
		f2.add_domains(cd)

		f1._propagate_domains_up()
		self.assertEqual(f1.domains, {})

	def test_domain_conflict(self):
		cda = ClockDomain('sync')
		cdb = ClockDomain('sync')

		fa = Fragment()
		fa.add_domains(cda)
		fb = Fragment()
		fb.add_domains(cdb)
		f = Fragment()
		f.add_subfragment(fa, 'a')
		f.add_subfragment(fb, 'b')

		f._propagate_domains_up()
		self.assertEqual(f.domains, {'a_sync': cda, 'b_sync': cdb})
		(fa, _), (fb, _) = f.subfragments
		self.assertEqual(fa.domains, {'a_sync': cda})
		self.assertEqual(fb.domains, {'b_sync': cdb})

	def test_domain_conflict_anon(self):
		cda = ClockDomain('sync')
		cdb = ClockDomain('sync')

		fa = Fragment()
		fa.add_domains(cda)
		fb = Fragment()
		fb.add_domains(cdb)
		f = Fragment()
		f.add_subfragment(fa, 'a')
		f.add_subfragment(fb)

		with self.assertRaisesRegex(
			DomainError, (
				r'^Domain \'sync\' is defined by subfragments \'a\', <unnamed #1> of fragment '
				r'\'top\'; it is necessary to either rename subfragment domains explicitly, '
				r'or give names to subfragments$'
			)
		):
			f._propagate_domains_up()

	def test_domain_conflict_name(self):
		cda = ClockDomain('sync')
		cdb = ClockDomain('sync')

		fa = Fragment()
		fa.add_domains(cda)
		fb = Fragment()
		fb.add_domains(cdb)
		f = Fragment()
		f.add_subfragment(fa, 'x')
		f.add_subfragment(fb, 'x')

		with self.assertRaisesRegex(
			DomainError, (
				r'^Domain \'sync\' is defined by subfragments #0, #1 of fragment \'top\', some '
				r'of which have identical names; it is necessary to either rename subfragment '
				r'domains explicitly, or give distinct names to subfragments$'
			)
		):
			f._propagate_domains_up()

	def test_domain_conflict_rename_drivers(self):
		cda = ClockDomain('sync')
		cdb = ClockDomain('sync')

		fa = Fragment()
		fa.add_domains(cda)
		fb = Fragment()
		fb.add_domains(cdb)
		fb.add_driver(ResetSignal('sync'), None)
		f = Fragment()
		f.add_subfragment(fa, 'a')
		f.add_subfragment(fb, 'b')

		f._propagate_domains_up()
		fb_new, _ = f.subfragments[1]
		self.assertEqual(fb_new.drivers, OrderedDict({
			None: SignalSet((ResetSignal('b_sync'),))
		}))

	def test_domain_conflict_rename_drivers_before_creating_missing(self):
		cda = ClockDomain('sync')
		cdb = ClockDomain('sync')
		s = Signal()

		fa = Fragment()
		fa.add_domains(cda)
		fb = Fragment()
		fb.add_domains(cdb)
		f = Fragment()
		f.add_subfragment(fa, 'a')
		f.add_subfragment(fb, 'b')
		f.add_driver(s, 'b_sync')

		f._propagate_domains(lambda name: ClockDomain(name))

	def test_propagate_down(self):
		cd = ClockDomain()

		f1 = Fragment()
		f2 = Fragment()
		f1.add_domains(cd)
		f1.add_subfragment(f2)

		f1._propagate_domains_down()
		self.assertEqual(f2.domains, {'cd': cd})

	def test_propagate_down_idempotent(self):
		cd = ClockDomain()

		f1 = Fragment()
		f1.add_domains(cd)
		f2 = Fragment()
		f2.add_domains(cd)
		f1.add_subfragment(f2)

		f1._propagate_domains_down()
		self.assertEqual(f1.domains, {'cd': cd})
		self.assertEqual(f2.domains, {'cd': cd})

	def test_propagate(self):
		cd = ClockDomain()

		f1 = Fragment()
		f2 = Fragment()
		f1.add_domains(cd)
		f1.add_subfragment(f2)

		new_domains = f1._propagate_domains(missing_domain = lambda name: None)
		self.assertEqual(f1.domains, {'cd': cd})
		self.assertEqual(f2.domains, {'cd': cd})
		self.assertEqual(new_domains, [])

	def test_propagate_missing(self):
		s1 = Signal()
		f1 = Fragment()
		f1.add_driver(s1, 'sync')

		with self.assertRaisesRegex(
			DomainError,
			r'^Domain \'sync\' is used but not defined$'
		):
			f1._propagate_domains(missing_domain = lambda name: None)

	def test_propagate_create_missing(self):
		s1 = Signal()
		f1 = Fragment()
		f1.add_driver(s1, 'sync')
		f2 = Fragment()
		f1.add_subfragment(f2)

		new_domains = f1._propagate_domains(missing_domain = lambda name: ClockDomain(name))
		self.assertEqual(f1.domains.keys(), {'sync'})
		self.assertEqual(f2.domains.keys(), {'sync'})
		self.assertEqual(f1.domains['sync'], f2.domains['sync'])
		self.assertEqual(new_domains, [f1.domains['sync']])

	def test_propagate_create_missing_fragment(self):
		s1 = Signal()
		f1 = Fragment()
		f1.add_driver(s1, 'sync')

		cd = ClockDomain('sync')
		f2 = Fragment()
		f2.add_domains(cd)

		new_domains = f1._propagate_domains(missing_domain = lambda name: f2)
		self.assertEqual(f1.domains.keys(), {'sync'})
		self.assertEqual(f1.domains['sync'], f2.domains['sync'])
		self.assertEqual(new_domains, [])
		self.assertEqual(f1.subfragments, [
			(f2, 'cd_sync')
		])

	def test_propagate_create_missing_fragment_many_domains(self):
		s1 = Signal()
		f1 = Fragment()
		f1.add_driver(s1, 'sync')

		cd_por  = ClockDomain('por')
		cd_sync = ClockDomain('sync')
		f2 = Fragment()
		f2.add_domains(cd_por, cd_sync)

		new_domains = f1._propagate_domains(missing_domain = lambda name: f2)
		self.assertEqual(f1.domains.keys(), {'sync', 'por'})
		self.assertEqual(f2.domains.keys(), {'sync', 'por'})
		self.assertEqual(f1.domains['sync'], f2.domains['sync'])
		self.assertEqual(new_domains, [])
		self.assertEqual(f1.subfragments, [
			(f2, 'cd_sync')
		])

	def test_propagate_create_missing_fragment_wrong(self):
		s1 = Signal()
		f1 = Fragment()
		f1.add_driver(s1, 'sync')

		f2 = Fragment()
		f2.add_domains(ClockDomain('foo'))

		with self.assertRaisesRegex(
			DomainError, (
				r'^Fragment returned by missing domain callback does not define requested '
				r'domain \'sync\' \(defines `foo`\)\.$'
			)
		):
			f1._propagate_domains(missing_domain = lambda name: f2)


class FragmentHierarchyConflictTestCase(ToriiTestSuiteCase):
	def setUp_self_sub(self):
		self.s1 = Signal()
		self.c1 = Signal()
		self.c2 = Signal()

		self.f1 = Fragment()
		self.f1.add_statements(self.c1.eq(0))
		self.f1.add_driver(self.s1)
		self.f1.add_driver(self.c1, 'sync')

		self.f1a = Fragment()
		self.f1.add_subfragment(self.f1a, 'f1a')

		self.f2 = Fragment()
		self.f2.add_statements(self.c2.eq(1))
		self.f2.add_driver(self.s1)
		self.f2.add_driver(self.c2, 'sync')
		self.f1.add_subfragment(self.f2)

		self.f1b = Fragment()
		self.f1.add_subfragment(self.f1b, 'f1b')

		self.f2a = Fragment()
		self.f2.add_subfragment(self.f2a, 'f2a')

	def test_conflict_self_sub(self):
		self.setUp_self_sub()

		self.f1._resolve_hierarchy_conflicts(mode = 'silent')
		self.assertEqual(self.f1.subfragments, [
			(self.f1a, 'f1a'),
			(self.f1b, 'f1b'),
			(self.f2a, 'f2a'),
		])
		self.assertRepr(self.f1.statements, '''
		(
			(eq (sig c1) (const 1'd0))
			(eq (sig c2) (const 1'd1))
		)
		''')
		self.assertEqual(self.f1.drivers, {
			None:   SignalSet((self.s1,)),
			'sync': SignalSet((self.c1, self.c2)),
		})

	def test_conflict_self_sub_error(self):
		self.setUp_self_sub()

		with self.assertRaisesRegex(
			DriverConflict,
			r'^Signal \'\(sig s1\)\' is driven from multiple fragments: top, top.<unnamed #1>$'
		):
			self.f1._resolve_hierarchy_conflicts(mode = 'error')

	def test_conflict_self_sub_warning(self):
		self.setUp_self_sub()

		with self.assertWarnsRegex(
			DriverConflict, (
				r'^Signal \'\(sig s1\)\' is driven from multiple fragments: top, top.<unnamed #1>; '
				r'hierarchy will be flattened$'
			)
		):
			self.f1._resolve_hierarchy_conflicts(mode = 'warn')

	def setUp_sub_sub(self):
		self.s1 = Signal()
		self.c1 = Signal()
		self.c2 = Signal()

		self.f1 = Fragment()

		self.f2 = Fragment()
		self.f2.add_driver(self.s1)
		self.f2.add_statements(self.c1.eq(0))
		self.f1.add_subfragment(self.f2)

		self.f3 = Fragment()
		self.f3.add_driver(self.s1)
		self.f3.add_statements(self.c2.eq(1))
		self.f1.add_subfragment(self.f3)

	def test_conflict_sub_sub(self):
		self.setUp_sub_sub()

		self.f1._resolve_hierarchy_conflicts(mode = 'silent')
		self.assertEqual(self.f1.subfragments, [])
		self.assertRepr(self.f1.statements, '''
		(
			(eq (sig c1) (const 1'd0))
			(eq (sig c2) (const 1'd1))
		)
		''')

	def setUp_self_subsub(self):
		self.s1 = Signal()
		self.c1 = Signal()
		self.c2 = Signal()

		self.f1 = Fragment()
		self.f1.add_driver(self.s1)

		self.f2 = Fragment()
		self.f2.add_statements(self.c1.eq(0))
		self.f1.add_subfragment(self.f2)

		self.f3 = Fragment()
		self.f3.add_driver(self.s1)
		self.f3.add_statements(self.c2.eq(1))
		self.f2.add_subfragment(self.f3)

	def test_conflict_self_subsub(self):
		self.setUp_self_subsub()

		self.f1._resolve_hierarchy_conflicts(mode = 'silent')
		self.assertEqual(self.f1.subfragments, [])
		self.assertRepr(self.f1.statements, '''
		(
			(eq (sig c1) (const 1'd0))
			(eq (sig c2) (const 1'd1))
		)
		''')

	def test_explicit_flatten(self):
		self.f1 = Fragment()
		self.f2 = Fragment()
		self.f2.flatten = True
		self.f1.add_subfragment(self.f2)

		self.f1._resolve_hierarchy_conflicts(mode = 'silent')
		self.assertEqual(self.f1.subfragments, [])

	def test_no_conflict_local_domains(self):
		f1 = Fragment()
		cd1 = ClockDomain('d', local = True)
		f1.add_domains(cd1)
		f1.add_driver(ClockSignal('d'))
		f2 = Fragment()
		cd2 = ClockDomain('d', local = True)
		f2.add_domains(cd2)
		f2.add_driver(ClockSignal('d'))
		f3 = Fragment()
		f3.add_subfragment(f1)
		f3.add_subfragment(f2)
		f3.prepare()


class InstanceTestCase(ToriiTestSuiteCase):
	def test_construct(self):
		s1 = Signal()
		s2 = Signal()
		s3 = Signal()
		s4 = Signal()
		s5 = Signal()
		s6 = Signal()
		inst = Instance('foo',
			('a', 'ATTR1', 1),
			('p', 'PARAM1', 0x1234),
			('i', 's1', s1),
			('o', 's2', s2),
			('io', 's3', s3),
			a_ATTR2 = 2,
			p_PARAM2 = 0x5678,
			i_s4 = s4,
			o_s5 = s5,
			io_s6 = s6,
		)
		self.assertEqual(inst.attrs, OrderedDict([
			('ATTR1', 1),
			('ATTR2', 2),
		]))
		self.assertEqual(inst.parameters, OrderedDict([
			('PARAM1', 0x1234),
			('PARAM2', 0x5678),
		]))
		self.assertEqual(inst.named_ports, OrderedDict([
			('s1', (s1, 'i')),
			('s2', (s2, 'o')),
			('s3', (s3, 'io')),
			('s4', (s4, 'i')),
			('s5', (s5, 'o')),
			('s6', (s6, 'io')),
		]))

	def test_cast_ports(self):
		inst = Instance('foo',
			('i', 's1', 1),
			('o', 's2', 2),
			('io', 's3', 3),
			i_s4 = 4,
			o_s5 = 5,
			io_s6 = 6,
		)
		self.assertRepr(inst.named_ports['s1'][0], '(const 1\'d1)')
		self.assertRepr(inst.named_ports['s2'][0], '(const 2\'d2)')
		self.assertRepr(inst.named_ports['s3'][0], '(const 2\'d3)')
		self.assertRepr(inst.named_ports['s4'][0], '(const 3\'d4)')
		self.assertRepr(inst.named_ports['s5'][0], '(const 3\'d5)')
		self.assertRepr(inst.named_ports['s6'][0], '(const 3\'d6)')

	def test_wrong_construct_arg(self):
		s = Signal()
		with self.assertRaisesRegex(
			NameError, (
				r'^Instance argument \(\'\', \'s1\', \(sig s\)\) should be a tuple '
				r'\(kind, name, value\) where kind is one of \'a\', \'p\', \'i\', \'o\', or \'io\'$'
			)
		):
			Instance('foo', ('', 's1', s))

	def test_wrong_construct_kwarg(self):
		s = Signal()
		with self.assertRaisesRegex(
			NameError, (
				r'^Instance keyword argument x_s1 = \(sig s\) does not start with one of '
				r'\'a_\', \'p_\', \'i_\', \'o_\', or \'io_\'$'
			)
		):
			Instance('foo', x_s1 = s)

	def setUp_cpu(self):
		self.rst = Signal()
		self.stb = Signal()
		self.pins = Signal(8)
		self.datal = Signal(4)
		self.datah = Signal(4)
		self.inst = Instance('cpu',
			p_RESET = 0x1234,
			i_clk = ClockSignal(),
			i_rst = self.rst,
			o_stb = self.stb,
			o_data = Cat(self.datal, self.datah),
			io_pins = self.pins[:]
		)
		self.wrap = Fragment()
		self.wrap.add_subfragment(self.inst)

	def test_init(self):
		self.setUp_cpu()
		f = self.inst
		self.assertEqual(f.type, 'cpu')
		self.assertEqual(f.parameters, OrderedDict([('RESET', 0x1234)]))
		self.assertEqual(list(f.named_ports.keys()), ['clk', 'rst', 'stb', 'data', 'pins'])
		self.assertEqual(f.ports, SignalDict([]))

	def test_prepare(self):
		self.setUp_cpu()
		f = self.wrap.prepare()
		sync_clk = f.domains['sync'].clk
		self.assertEqual(f.ports, SignalDict([
			(sync_clk, 'i'),
			(self.rst, 'i'),
			(self.pins, 'io'),
		]))

	def test_prepare_explicit_ports(self):
		self.setUp_cpu()
		f = self.wrap.prepare(ports = [self.rst, self.stb])
		sync_clk = f.domains['sync'].clk
		sync_rst = f.domains['sync'].rst
		self.assertEqual(f.ports, SignalDict([
			(sync_clk, 'i'),
			(sync_rst, 'i'),
			(self.rst, 'i'),
			(self.stb, 'o'),
			(self.pins, 'io'),
		]))

	def test_prepare_slice_in_port(self):
		s = Signal(2)
		f = Fragment()
		f.add_subfragment(Instance('foo', o_O = s[0]))
		f.add_subfragment(Instance('foo', o_O = s[1]))
		fp = f.prepare(ports = [s], missing_domain = lambda name: None)
		self.assertEqual(fp.ports, SignalDict([
			(s, 'o'),
		]))

	def test_prepare_attrs(self):
		self.setUp_cpu()
		self.inst.attrs['ATTR'] = 1
		f = self.inst.prepare()
		self.assertEqual(f.attrs, OrderedDict([
			('ATTR', 1),
		]))



# SPDX-License-Identifier: BSD-2-Clause

import hashlib
import os
import subprocess
import sys
import tempfile
import zipfile
from abc         import ABCMeta, abstractmethod
from collections import OrderedDict
from contextlib  import contextmanager
from pathlib     import Path
from typing      import Generator, Literal, Union

__all__ = (
	'BuildPlan',
	'BuildProducts',
	'LocalBuildProducts',
)

class BuildPlan:
	def __init__(self, script: str) -> None:
		'''
		A build plan.

		Parameters
		----------
		script : str
			The base name (without extension) of the script that will be executed.

		'''
		self.script = script
		self.files  = OrderedDict()

	def add_file(self, filename: str, content: Union[str, bytes]) -> None:
		'''
		Add ``content``, which can be a :class:`str`` or :class:`bytes`, to the build plan
		as ``filename``. The file name can be a relative path with directories separated by
		forward slashes (``/``).
		'''
		assert isinstance(filename, str) and filename not in self.files
		if Path(filename).is_absolute():
			raise ValueError(f'Filename {filename} must not be an absolute path')

		self.files[filename] = content

	def digest(self, size: int = 64) -> bytes:
		'''
		Compute a `digest`, a short byte sequence deterministically and uniquely identifying
		this build plan.
		'''
		hasher = hashlib.blake2b(digest_size = size)
		for filename in sorted(self.files):
			hasher.update(filename.encode('utf-8'))
			content = self.files[filename]
			if isinstance(content, str):
				content = content.encode('utf-8')
			hasher.update(content)
		hasher.update(self.script.encode('utf-8'))
		return hasher.digest()

	def archive(self, file: str) -> None:
		'''
		Archive files from the build plan into ``file``, which can be either a filename, or
		a file-like object. The produced archive is deterministic: exact same files will
		always produce exact same archive.
		'''
		with zipfile.ZipFile(file, 'w') as archive:
			# Write archive members in deterministic order and with deterministic timestamp.
			for filename in sorted(self.files):
				archive.writestr(zipfile.ZipInfo(filename), self.files[filename])

	def execute_local(
		self, root: Union[str, Path] = 'build', *, run_script: bool = True, env: dict[str, str] = None
	) -> 'LocalBuildProducts':
		'''
		Execute build plan using the local strategy. Files from the build plan are placed in
		the build root directory ``root``, and, if ``run_script`` is ``True``, the script
		appropriate for the platform (``{script}.bat`` on Windows, ``{script}.sh`` elsewhere) is
		executed in the build root. If ``env`` is not ``None``, the environment is extended
		(not replaced) with ``env``.

		Returns :class:`LocalBuildProducts`.
		'''

		if isinstance(root, str):
			root = Path(root).resolve()

		root.mkdir(parents = True, exist_ok = True)

		cwd = Path.cwd()
		try:
			os.chdir(root)

			for filename, content in self.files.items():
				filename = Path(filename)
				# Forbid parent directory components and absolute paths completely
				# to avoid the possibility of writing outside the build root.
				if '..' in filename.parts or filename.is_absolute():
					raise RuntimeError(
						f'Unable to write to \'{filename}\'\n'
						'Writing to outside of the build root is forbidden.'
					)

				filename.parent.mkdir(parents = True, exist_ok = True)

				if isinstance(content, str):
					content = content.encode('utf-8')

				with filename.resolve().open('wb') as f:
					f.write(content)

			if run_script:
				script_env = dict(os.environ)
				if env is not None:
					script_env.update(env)
				if sys.platform.startswith('win32'):
					# Without "call", "cmd /c {}.bat" will return 0.
					# See https://stackoverflow.com/a/30736987 for a detailed explanation of why.
					# Running the script manually from a command prompt is unaffected.
					subprocess.check_call(
						[ 'cmd', '/c', f'call {self.script}.bat' ],
						env = script_env
					)
				else:
					subprocess.check_call(
						[ 'sh', f'{self.script}.sh' ],
						env = script_env
					)

			return LocalBuildProducts(Path.cwd())

		finally:
			os.chdir(cwd)

	def execute(self) -> 'LocalBuildProducts':
		'''
		Execute build plan using the default strategy. Use one of the ``execute_*`` methods
		explicitly to have more control over the strategy.
		'''
		return self.execute_local()


class BuildProducts(metaclass = ABCMeta):
	@abstractmethod
	def get(self, filename: str, mode: Literal['b', 't'] = 'b') -> Union[str, bytes]:
		'''
		Extract ``filename`` from build products, and return it as a :class:`bytes` (if ``mode``
		is ``"b"``) or a :class:`str` (if ``mode`` is ``"t"``).
		'''
		if mode not in ('b', 't'):
			raise ValueError(f'Unsupported file access mode \'{mode}\', must be either \'b\' or \'t\'.')


	@contextmanager
	def extract(self, *filenames: tuple[str]) -> Generator[
		Union[None, str, list[str]], None, None
	]:
		'''
		Extract ``filenames`` from build products, place them in an OS-specific temporary file
		location, with the extension preserved, and delete them afterwards. This method is used
		as a context manager, e.g.: ::

			with products.extract("bitstream.bin", "programmer.cfg") \
					as bitstream_filename, config_filename:
				subprocess.check_call(["program", "-c", config_filename, bitstream_filename])
		'''
		files = []
		try:
			for filename in filenames:
				# On Windows, a named temporary file (as created by Python) is not accessible to
				# others if it's still open within the Python process, so we close it and delete
				# it manually.
				file = tempfile.NamedTemporaryFile(
					prefix = 'torii_', suffix = '_' + os.path.basename(filename),
					delete = False)
				files.append(file)
				file.write(self.get(filename))
				file.close()

			if len(files) == 0:
				return (yield)
			elif len(files) == 1:
				return (yield files[0].name)
			else:
				return (yield [file.name for file in files])
		finally:
			for file in files:
				os.unlink(file.name)


class LocalBuildProducts(BuildProducts):
	def __init__(self, root: Union[str, Path]) -> None:
		# We provide no guarantees that files will be available on the local filesystem (i.e. in
		# any way other than through `products.get()`) in general, so downstream code must never
		# rely on this, even when we happen to use a local build most of the time.
		if isinstance(root, str):
			self.__root = Path(root)
		else:
			self.__root = root

	def get(self, filename: str, mode: Literal['b', 't'] = 'b') -> Union[str, bytes]:
		super().get(filename, mode)
		with (self.__root / filename).resolve().open(f'r{mode}') as f:
			return f.read()



# SPDX-License-Identifier: BSD-2-Clause

from torii.hdl     import *
from torii.hdl.rec import *
from torii.sim     import *
from torii.lib.io  import *

from ..utils       import ToriiTestSuiteCase


class PinLayoutTestCase(ToriiTestSuiteCase):
	def assertLayoutEqual(self, layout, expected):
		casted_layout = {}
		for name, (shape, dir) in layout.items():
			casted_layout[name] = (Shape.cast(shape), dir)

		self.assertEqual(casted_layout, expected)


class PinLayoutCombTestCase(PinLayoutTestCase):
	def test_pin_layout_i(self):
		layout_1 = pin_layout(1, dir = 'i')
		self.assertLayoutEqual(layout_1.fields, {
			'i': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'i')
		self.assertLayoutEqual(layout_2.fields, {
			'i': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_o(self):
		layout_1 = pin_layout(1, dir = 'o')
		self.assertLayoutEqual(layout_1.fields, {
			'o': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'o')
		self.assertLayoutEqual(layout_2.fields, {
			'o': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_oe(self):
		layout_1 = pin_layout(1, dir = 'oe')
		self.assertLayoutEqual(layout_1.fields, {
			'o':  (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'oe')
		self.assertLayoutEqual(layout_2.fields, {
			'o':  (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

	def test_pin_layout_io(self):
		layout_1 = pin_layout(1, dir = 'io')
		self.assertLayoutEqual(layout_1.fields, {
			'i':  (unsigned(1), DIR_NONE),
			'o':  (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'io')
		self.assertLayoutEqual(layout_2.fields, {
			'i':  (unsigned(2), DIR_NONE),
			'o':  (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})


class PinLayoutSDRTestCase(PinLayoutTestCase):
	def test_pin_layout_i(self):
		layout_1 = pin_layout(1, dir = 'i', xdr = 1)
		self.assertLayoutEqual(layout_1.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'i', xdr = 1)
		self.assertLayoutEqual(layout_2.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_o(self):
		layout_1 = pin_layout(1, dir = 'o', xdr = 1)
		self.assertLayoutEqual(layout_1.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'o', xdr = 1)
		self.assertLayoutEqual(layout_2.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_oe(self):
		layout_1 = pin_layout(1, dir = 'oe', xdr = 1)
		self.assertLayoutEqual(layout_1.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o':  (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'oe', xdr = 1)
		self.assertLayoutEqual(layout_2.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o':  (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

	def test_pin_layout_io(self):
		layout_1 = pin_layout(1, dir = 'io', xdr = 1)
		self.assertLayoutEqual(layout_1.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i':  (unsigned(1), DIR_NONE),
			'o_clk': (unsigned(1), DIR_NONE),
			'o':  (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'io', xdr = 1)
		self.assertLayoutEqual(layout_2.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i':  (unsigned(2), DIR_NONE),
			'o_clk': (unsigned(1), DIR_NONE),
			'o':  (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})


class PinLayoutDDRTestCase(PinLayoutTestCase):
	def test_pin_layout_i(self):
		layout_1 = pin_layout(1, dir = 'i', xdr = 2)
		self.assertLayoutEqual(layout_1.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i0': (unsigned(1), DIR_NONE),
			'i1': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'i', xdr = 2)
		self.assertLayoutEqual(layout_2.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i0': (unsigned(2), DIR_NONE),
			'i1': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_o(self):
		layout_1 = pin_layout(1, dir = 'o', xdr = 2)
		self.assertLayoutEqual(layout_1.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(1), DIR_NONE),
			'o1': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'o', xdr = 2)
		self.assertLayoutEqual(layout_2.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(2), DIR_NONE),
			'o1': (unsigned(2), DIR_NONE),
		})

	def test_pin_layout_oe(self):
		layout_1 = pin_layout(1, dir = 'oe', xdr = 2)
		self.assertLayoutEqual(layout_1.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(1), DIR_NONE),
			'o1': (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'oe', xdr = 2)
		self.assertLayoutEqual(layout_2.fields, {
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(2), DIR_NONE),
			'o1': (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

	def test_pin_layout_io(self):
		layout_1 = pin_layout(1, dir = 'io', xdr = 2)
		self.assertLayoutEqual(layout_1.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i0': (unsigned(1), DIR_NONE),
			'i1': (unsigned(1), DIR_NONE),
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(1), DIR_NONE),
			'o1': (unsigned(1), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})

		layout_2 = pin_layout(2, dir = 'io', xdr = 2)
		self.assertLayoutEqual(layout_2.fields, {
			'i_clk': (unsigned(1), DIR_NONE),
			'i0': (unsigned(2), DIR_NONE),
			'i1': (unsigned(2), DIR_NONE),
			'o_clk': (unsigned(1), DIR_NONE),
			'o0': (unsigned(2), DIR_NONE),
			'o1': (unsigned(2), DIR_NONE),
			'oe': (unsigned(1), DIR_NONE),
		})


class PinTestCase(ToriiTestSuiteCase):
	def test_attributes(self):
		pin = Pin(2, dir = 'io', xdr = 2)
		self.assertEqual(pin.width, 2)
		self.assertEqual(pin.dir,   'io')
		self.assertEqual(pin.xdr,   2)



# SPDX-License-Identifier: BSD-2-Clause

from collections import OrderedDict
from typing      import Callable, Generator, Iterator, Literal, Optional, Union

__all__ = (
	'Attrs',
	'Clock',
	'Connector',
	'DiffPairs',
	'DiffPairsN',
	'Pins',
	'PinsN',
	'Resource',
	'Subsignal',
)

class Pins:
	def __init__(
		self, names: str, *, dir: Literal['i', 'o', 'io', 'oe'] = 'io', invert: bool = False,
		conn: Optional[tuple[str, Union[int, str]]] = None, assert_width: Optional[int] = None
	) -> None:
		if not isinstance(names, str):
			raise TypeError(f'Names must be a whitespace-separated string, not {names!r}')
		names = names.split()

		if conn is not None:
			conn_name, conn_number = conn
			if not (isinstance(conn_name, str) and isinstance(conn_number, (int, str))):
				raise TypeError(
					'Connector must be None or a pair of string (connector name) and '
					f'integer/string (connector number), not {conn!r}'
				)
			names = [ f'{conn_name}_{conn_number}:{name}' for name in names ]

		if dir not in ('i', 'o', 'io', 'oe'):
			raise TypeError(f'Direction must be one of \'i\', \'o\', \'oe\', or \'io\', not {dir!r}')

		if assert_width is not None and len(names) != assert_width:
			raise AssertionError(f'{len(names)} names are specified ({" ".join(names)}), but {assert_width} names are expected')

		self.names  = names
		self.dir    = dir
		self.invert = bool(invert)

	def __len__(self) -> int:
		return len(self.names)

	def __iter__(self) -> Iterator[str]:
		return iter(self.names)

	def map_names(self, mapping: dict[str, str], resource) -> list[str]:
		mapped_names = []
		for name in self.names:
			while ":" in name:
				if name not in mapping:
					raise NameError(f'Resource {resource!r} refers to nonexistent connector pin {name}')
				name = mapping[name]
			mapped_names.append(name)
		return mapped_names

	def __repr__(self) -> str:
		return f'(pins{"-n" if self.invert else ""} {self.dir} {" ".join(self.names)})'


def PinsN(*args, **kwargs) -> Pins:
	return Pins(*args, invert = True, **kwargs)


class DiffPairs:
	def __init__(
		self, p: str, n: str, *, dir: Literal['i', 'o', 'io', 'oe'] = 'io', invert: bool = False,
		conn: Optional[tuple[str, Union[int, str]]] = None, assert_width: Optional[bool] = None
	) -> None:
		self.p = Pins(p, dir = dir, conn = conn, assert_width = assert_width)
		self.n = Pins(n, dir = dir, conn = conn, assert_width = assert_width)

		if len(self.p.names) != len(self.n.names):
			raise TypeError(f'Positive and negative pins must have the same width, but {self.p!r} '
							f'and {self.n!r} do not')

		self.dir    = dir
		self.invert = bool(invert)

	def __len__(self) -> int:
		return len(self.p.names)

	def __iter__(self) -> Iterator[tuple[str, str]]:
		return zip(self.p.names, self.n.names)

	def __repr__(self) -> str:
		return f'(diffpairs{"-n" if self.invert else ""} {self.dir} (p {" ".join(self.p.names)}) (n {" ".join(self.n.names)}))'


def DiffPairsN(*args, **kwargs) -> DiffPairs:
	return DiffPairs(*args, invert = True, **kwargs)


class Attrs(OrderedDict):
	def __init__(self, **attrs: dict[str, Union[int, str, Callable]]) -> None:
		for key, value in attrs.items():
			if not (value is None or isinstance(value, (str, int)) or hasattr(value, '__call__')):
				raise TypeError(f'Value of attribute {key} must be None, int, str, or callable, not {value!r}')

		super().__init__(**attrs)

	def __repr__(self) -> str:
		items = []
		for key, value in self.items():
			if value is None:
				items.append('!' + key)
			else:
				items.append(key + '=' + repr(value))
		return f'(attrs {" ".join(items)})'


class Clock:
	def __init__(self, frequency: Union[float, int]) -> None:
		if not isinstance(frequency, (float, int)):
			raise TypeError('Clock frequency must be a number')

		self.frequency = float(frequency)

	@property
	def period(self) -> float:
		return 1 / self.frequency

	def __repr__(self) -> str:
		return f'(clock {self.frequency})'


class Subsignal:
	def __init__(
		self, name: str, *args: Union[Pins, DiffPairs, 'Subsignal', Attrs, Clock]
	) -> None:
		self.name  = name
		self.ios   = []
		self.attrs = Attrs()
		self.clock = None

		if not args:
			raise ValueError('Missing I/O constraints')
		for arg in args:
			if isinstance(arg, (Pins, DiffPairs)):
				if not self.ios:
					self.ios.append(arg)
				else:
					raise TypeError(
						'Pins and DiffPairs are incompatible with other location or '
						f'subsignal constraints, but {arg!r} appears after {self.ios[-1]!r}'
					)

			elif isinstance(arg, Subsignal):
				if not self.ios or isinstance(self.ios[-1], Subsignal):
					self.ios.append(arg)
				else:
					raise TypeError(
						'Subsignal is incompatible with location constraints, but '
						f'{arg!r} appears after {self.ios[-1]!r}'
					)
			elif isinstance(arg, Attrs):
				self.attrs.update(arg)
			elif isinstance(arg, Clock):
				if self.ios and isinstance(self.ios[-1], (Pins, DiffPairs)):
					if self.clock is None:
						self.clock = arg
					else:
						raise ValueError('Clock constraint can be applied only once')
				else:
					raise TypeError(f'Clock constraint can only be applied to Pins or DiffPairs, not {self.ios[-1]!r}')
			else:
				raise TypeError(f'Constraint must be one of Pins, DiffPairs, Subsignal, Attrs, or Clock, not {arg!r}')

	def _content_repr(self) -> str:
		parts = []
		for io in self.ios:
			parts.append(repr(io))
		if self.clock is not None:
			parts.append(repr(self.clock))
		if self.attrs:
			parts.append(repr(self.attrs))
		return " ".join(parts)

	def __repr__(self) -> str:
		return f'(subsignal {self.name} {self._content_repr()})'

class Resource(Subsignal):
	@classmethod
	def family(
		cls, name_or_number: Union[str, int], number: Optional[int] = None, *,
		ios: list[Union[Pins, DiffPairs, 'Subsignal', Attrs, Clock]], default_name: str, name_suffix: str = ''
	) -> 'Resource':
		# This constructor accepts two different forms:
		#  1. Number-only form:
		#       Resource.family(0, default_name = 'name', ios = [ Pins('A0 A1') ])
		#  2. Name-and-number (name override) form:
		#       Resource.family('override', 0, default_name = "name", ios = ...)
		# This makes it easier to build abstractions for resources, e.g. an SPIResource abstraction
		# could simply delegate to `Resource.family(*args, default_name = 'spi', ios = ios)`.
		# The name_suffix argument is meant to support creating resources with
		# similar names, such as spi_flash, spi_flash_2x, etc.
		if name_suffix:  # Only add '_' if we actually have a suffix.
			name_suffix = '_' + name_suffix

		if number is None: # name_or_number is number
			return cls(default_name + name_suffix, name_or_number, *ios)
		else: # name_or_number is name
			return cls(name_or_number + name_suffix, number, *ios)

	def __init__(
		self, name: str, number: int, *args: Union[Pins, DiffPairs, 'Subsignal', Attrs, Clock]
	) -> None:
		if not isinstance(number, int):
			raise TypeError(f'Resource number must be an integer, not {number!r}')

		super().__init__(name, *args)
		self.number = number

	def __repr__(self) -> str:
		return f'(resource {self.name} {self.number} {self._content_repr()})'

class Connector:
	def __init__(
		self, name: str, number: int, io: Union[str, dict[str, str]], *, conn: Optional[tuple[str, Union[int, str]]] = None
	) -> None:
		self.name    = name
		self.number  = number
		mapping = OrderedDict()

		if isinstance(io, dict):
			for conn_pin, plat_pin in io.items():
				if not isinstance(conn_pin, str):
					raise TypeError(f'Connector pin name must be a string, not {conn_pin!r}')
				if not isinstance(plat_pin, str):
					raise TypeError(f'Platform pin name must be a string, not {plat_pin!r}')
				mapping[conn_pin] = plat_pin

		elif isinstance(io, str):
			for conn_pin, plat_pin in enumerate(io.split(), start = 1):
				if plat_pin == '-':
					continue

				mapping[str(conn_pin)] = plat_pin
		else:
			raise TypeError(f'Connector I/Os must be a dictionary or a string, not {io!r}')

		if conn is not None:
			conn_name, conn_number = conn
			if not (isinstance(conn_name, str) and isinstance(conn_number, (int, str))):
				raise TypeError('Connector must be None or a pair of string (connector name) and '
								f'integer/string (connector number), not {conn!r}')

			for conn_pin, plat_pin in mapping.items():
				mapping[conn_pin] = f'{conn_name}_{conn_number}:{plat_pin}'
		self.mapping = mapping

	def __repr__(self) -> str:
		return f'(connector {self.name} {self.number} {" ".join(f"{conn}=>{plat}" for conn, plat in self.mapping.items())})'

	def __len__(self) -> int:
		return len(self.mapping)

	def __iter__(self) -> Generator[tuple[str, str], None, None]:
		for conn_pin, plat_pin in self.mapping.items():
			yield f'{self.name}_{self.number}:{conn_pin}', plat_pin



# SPDX-License-Identifier: BSD-2-Clause

import warnings
from collections  import OrderedDict
from contextlib   import _GeneratorContextManager, contextmanager
from enum         import Enum
from functools    import wraps
from sys          import version_info

from ..util       import flatten, tracer
from ..util.units import bits_for
from .ast         import (
	Assert, Assign, Assume, Cat, Cover, Operator, Signal, SignalDict,
	Statement, Switch, Value
)
from .cd          import ClockDomain
from .ir          import Elaboratable, Fragment
from .xfrm        import SampleDomainInjector

__all__ = (
	'Module',
	'SyntaxError',
	'SyntaxWarning',
)


class SyntaxError(Exception):
	pass


class SyntaxWarning(Warning):
	pass


class _ModuleBuilderProxy:
	def __init__(self, builder, depth):
		object.__setattr__(self, '_builder', builder)
		object.__setattr__(self, '_depth', depth)


class _ModuleBuilderDomain(_ModuleBuilderProxy):
	def __init__(self, builder, depth, domain):
		super().__init__(builder, depth)
		self._domain = domain

	def __iadd__(self, assigns):
		self._builder._add_statement(assigns, domain = self._domain, depth = self._depth)
		return self


class _ModuleBuilderDomains(_ModuleBuilderProxy):
	def __getattr__(self, name):
		if name == 'submodules':
			warnings.warn(
				f'Using \'<module>.d.{name}\' would add statements to clock domain {name!r}; '
				f'did you mean <module>.{name} instead?',
				SyntaxWarning, stacklevel = 2
			)
		if name == 'comb':
			domain = None
		else:
			domain = name
		return _ModuleBuilderDomain(self._builder, self._depth, domain)

	def __getitem__(self, name):
		return self.__getattr__(name)

	def __setattr__(self, name, value):
		if name == '_depth':
			object.__setattr__(self, name, value)
		elif not isinstance(value, _ModuleBuilderDomain):
			raise AttributeError(f'Cannot assign \'d.{name}\' attribute; did you mean \'d.{name} +=\'?')

	def __setitem__(self, name, value):
		return self.__setattr__(name, value)


class _ModuleBuilderRoot:
	def __init__(self, builder, depth):
		self._builder = builder
		self.domain = self.d = _ModuleBuilderDomains(builder, depth)

	def __getattr__(self, name):
		if name in ('comb', 'sync'):
			raise AttributeError(f'\'{type(self).__name__}\' object has no attribute \'{name}\'; did you mean \'d.{name}\'?')
		raise AttributeError(f'\'{type(self).__name__}\' object has no attribute \'{name}\'')


class _ModuleBuilderSubmodules:
	def __init__(self, builder):
		object.__setattr__(self, '_builder', builder)

	def __iadd__(self, modules):
		for module in flatten([modules]):
			self._builder._add_submodule(module)
		return self

	def __setattr__(self, name, submodule):
		self._builder._add_submodule(submodule, name)

	def __setitem__(self, name, value):
		return self.__setattr__(name, value)

	def __getattr__(self, name):
		return self._builder._get_submodule(name)

	def __getitem__(self, name):
		return self.__getattr__(name)


class _ModuleBuilderDomainSet:
	def __init__(self, builder):
		object.__setattr__(self, '_builder', builder)

	def __iadd__(self, domains):
		for domain in flatten([domains]):
			if not isinstance(domain, ClockDomain):
				raise TypeError(f'Only clock domains may be added to `m.domains`, not {domain!r}')
			self._builder._add_domain(domain)
		return self

	def __setattr__(self, name, domain):
		if not isinstance(domain, ClockDomain):
			raise TypeError(f'Only clock domains may be added to `m.domains`, not {domain!r}')
		if domain.name != name:
			raise NameError(f'Clock domain name {domain.name!r} must match name in `m.domains.{name} += ...` syntax')
		self._builder._add_domain(domain)


# It's not particularly clean to depend on an internal interface, but, unfortunately, __bool__
# must be defined on a class to be called during implicit conversion.
class _GuardedContextManager(_GeneratorContextManager):
	def __init__(self, keyword, func, args, kwds):
		self.keyword = keyword
		return super().__init__(func, args, kwds)

	def __bool__(self):
		raise SyntaxError(f'`if m.{self.keyword}(...):` does not work; use `with m.{self.keyword}(...)`')


def _guardedcontextmanager(keyword):
	def decorator(func):
		@wraps(func)
		def helper(*args, **kwds):
			return _GuardedContextManager(keyword, func, args, kwds)
		return helper
	return decorator


class FSM:
	def __init__(self, state, encoding, decoding):
		self.state    = state
		self.encoding = encoding
		self.decoding = decoding

	def ongoing(self, name):
		if name not in self.encoding:
			self.encoding[name] = len(self.encoding)
		return Operator('==', [ self.state, self.encoding[name] ], src_loc_at = 0)


class Module(_ModuleBuilderRoot, Elaboratable):
	@classmethod
	def __init_subclass__(cls):
		raise SyntaxError(
			'Instead of inheriting from `Module`, inherit from `Elaboratable` '
			'and return a `Module` from the `elaborate(self, platform)` method'
		)

	def __init__(self):
		_ModuleBuilderRoot.__init__(self, self, depth = 0)
		self.submodules    = _ModuleBuilderSubmodules(self)
		self.domains       = _ModuleBuilderDomainSet(self)

		self._statements   = Statement.cast([])
		self._ctrl_context = None
		self._ctrl_stack   = []

		self._driving      = SignalDict()
		self._named_submodules = {}
		self._anon_submodules  = []
		self._domains      = {}
		self._generated    = {}

	def _check_context(self, construct, context):
		if self._ctrl_context != context:
			if self._ctrl_context is None:
				raise SyntaxError(f'{construct} is not permitted outside of {context}')
			else:
				if self._ctrl_context == 'Switch':
					secondary_context = 'Case'
				if self._ctrl_context == 'FSM':
					secondary_context = 'State'
				raise SyntaxError(
					f'{construct} is not permitted directly inside of {self._ctrl_context}; '
					f'it is permitted inside of {self._ctrl_context} {secondary_context}'
				)

	def _get_ctrl(self, name):
		if self._ctrl_stack:
			top_name, top_data = self._ctrl_stack[-1]
			if top_name == name:
				return top_data

	def _flush_ctrl(self):
		while len(self._ctrl_stack) > self.domain._depth:
			self._pop_ctrl()

	def _set_ctrl(self, name, data):
		self._flush_ctrl()
		self._ctrl_stack.append((name, data))
		return data

	def _check_signed_cond(self, cond):
		cond = Value.cast(cond)
		if version_info < (3, 12, 0) and cond.shape().signed:
			# TODO(py3.11): remove; ~True is a warning in 3.12+, finally!
			warnings.warn(
				'Signed values in If/Elif conditions usually result from inverting '
				'Python booleans with ~, which leads to unexpected results. '
				'Replace `~flag` with `not flag`. (If this is a false positive, '
				'silence this warning with `m.If(x)` → `m.If(x.bool())`.)',
				SyntaxWarning, stacklevel = 4
			)
		return cond

	@_guardedcontextmanager('If')
	def If(self, cond):
		self._check_context('If', context = None)
		cond = self._check_signed_cond(cond)
		src_loc = tracer.get_src_loc(src_loc_at = 1)
		if_data = self._set_ctrl('If', {
			'depth'   : self.domain._depth,
			'tests'   : [],
			'bodies'  : [],
			'src_loc' : src_loc,
			'src_locs': [],
		})
		try:
			_outer_case, self._statements = self._statements, []
			self.domain._depth += 1
			yield
			self._flush_ctrl()
			if_data['tests'].append(cond)
			if_data['bodies'].append(self._statements)
			if_data['src_locs'].append(src_loc)
		finally:
			self.domain._depth -= 1
			self._statements = _outer_case

	@_guardedcontextmanager('Elif')
	def Elif(self, cond):
		self._check_context('Elif', context = None)
		cond = self._check_signed_cond(cond)
		src_loc = tracer.get_src_loc(src_loc_at = 1)
		if_data = self._get_ctrl('If')
		if if_data is None or if_data['depth'] != self.domain._depth:
			raise SyntaxError('Elif without preceding If')
		try:
			_outer_case, self._statements = self._statements, []
			self.domain._depth += 1
			yield
			self._flush_ctrl()
			if_data['tests'].append(cond)
			if_data['bodies'].append(self._statements)
			if_data['src_locs'].append(src_loc)
		finally:
			self.domain._depth -= 1
			self._statements = _outer_case

	@_guardedcontextmanager('Else')
	def Else(self):
		self._check_context('Else', context = None)
		src_loc = tracer.get_src_loc(src_loc_at = 1)
		if_data = self._get_ctrl('If')
		if if_data is None or if_data['depth'] != self.domain._depth:
			raise SyntaxError('Else without preceding If/Elif')
		try:
			_outer_case, self._statements = self._statements, []
			self.domain._depth += 1
			yield
			self._flush_ctrl()
			if_data['bodies'].append(self._statements)
			if_data['src_locs'].append(src_loc)
		finally:
			self.domain._depth -= 1
			self._statements = _outer_case
		self._pop_ctrl()

	@contextmanager
	def Switch(self, test):
		self._check_context('Switch', context = None)
		self._set_ctrl('Switch', {
			'test'         : Value.cast(test),
			'cases'        : OrderedDict(),
			'src_loc'      : tracer.get_src_loc(src_loc_at = 1),
			'case_src_locs': {},
		})
		try:
			self._ctrl_context = 'Switch'
			self.domain._depth += 1
			yield
		finally:
			self.domain._depth -= 1
			self._ctrl_context = None
		self._pop_ctrl()

	@contextmanager
	def Case(self, *patterns):
		self._check_context('Case', context = 'Switch')
		src_loc = tracer.get_src_loc(src_loc_at = 1)
		switch_data = self._get_ctrl('Switch')
		new_patterns = ()
		for pattern in patterns:
			if not isinstance(pattern, (int, str, Enum)):
				raise SyntaxError(f'Case pattern must be an integer, a string, or an enumeration, not {pattern!r}')
			if isinstance(pattern, str) and any(bit not in '01- \t' for bit in pattern):
				raise SyntaxError(
					f'Case pattern \'{pattern}\' must consist of 0, 1, and - (don\'t care) bits, and may '
					'include whitespace'
				)
			if (isinstance(pattern, str) and
					len("".join(pattern.split())) != len(switch_data['test'])):
				raise SyntaxError(
					f'Case pattern \'{pattern}\' must have the same width as switch value '
					f'(which is {len(switch_data["test"])})'
				)
			if isinstance(pattern, int) and bits_for(pattern) > len(switch_data["test"]):
				warnings.warn(
					f'Case pattern \'{pattern:b}\' is wider than switch value (which has width {len(switch_data["test"])}); '
					'comparison will never be true', SyntaxWarning, stacklevel = 3
				)
				continue
			if isinstance(pattern, Enum) and bits_for(pattern.value) > len(switch_data["test"]):
				warnings.warn(
					f'Case pattern \'{pattern.value:b}\' ({ pattern.__class__.__name__}.{pattern.name}) is wider than switch value '
					f'(which has width {len(switch_data["test"])}); comparison will never be true',
					SyntaxWarning, stacklevel = 3
				)
				continue
			new_patterns = (*new_patterns, pattern)
		try:
			_outer_case, self._statements = self._statements, []
			self._ctrl_context = None
			yield
			self._flush_ctrl()
			# If none of the provided cases can possibly be true, omit this branch completely.
			# This needs to be differentiated from no cases being provided in the first place,
			# which means the branch will always match.
			if not (patterns and not new_patterns):
				switch_data['cases'][new_patterns] = self._statements
				switch_data['case_src_locs'][new_patterns] = src_loc
		finally:
			self._ctrl_context = 'Switch'
			self._statements = _outer_case

	def Default(self):
		return self.Case()

	@contextmanager
	def FSM(self, reset = None, domain = 'sync', name = 'fsm'):
		self._check_context('FSM', context = None)
		if domain == 'comb':
			raise ValueError(f'FSM may not be driven by the \'{domain}\' domain')
		fsm_data = self._set_ctrl('FSM', {
			'name'          : name,
			'signal'        : Signal(name = f'{name}_state', src_loc_at = 2),
			'reset'         : reset,
			'domain'        : domain,
			'encoding'      : OrderedDict(),
			'decoding'      : OrderedDict(),
			'states'        : OrderedDict(),
			'src_loc'       : tracer.get_src_loc(src_loc_at = 1),
			'state_src_locs': {},
		})
		self._generated[name] = fsm = FSM(
			fsm_data['signal'], fsm_data['encoding'], fsm_data['decoding']
		)
		try:
			self._ctrl_context = 'FSM'
			self.domain._depth += 1
			yield fsm
			for state_name in fsm_data['encoding']:
				if state_name not in fsm_data['states']:
					raise NameError(f'FSM state \'{state_name}\' is referenced but not defined')
		finally:
			self.domain._depth -= 1
			self._ctrl_context = None
		self._pop_ctrl()

	@contextmanager
	def State(self, name):
		self._check_context('FSM State', context = 'FSM')
		src_loc = tracer.get_src_loc(src_loc_at = 1)
		fsm_data = self._get_ctrl('FSM')
		if name in fsm_data['states']:
			raise NameError(f'FSM state \'{name}\' is already defined')
		if name not in fsm_data['encoding']:
			fsm_data['encoding'][name] = len(fsm_data['encoding'])
		try:
			_outer_case, self._statements = self._statements, []
			self._ctrl_context = None
			yield
			self._flush_ctrl()
			fsm_data['states'][name] = self._statements
			fsm_data['state_src_locs'][name] = src_loc
		finally:
			self._ctrl_context = 'FSM'
			self._statements = _outer_case

	@property
	def next(self):
		raise SyntaxError('Only assignment to `m.next` is permitted')

	@next.setter
	def next(self, name):
		if self._ctrl_context != 'FSM':
			for level, (ctrl_name, ctrl_data) in enumerate(reversed(self._ctrl_stack)):
				if ctrl_name == 'FSM':
					if name not in ctrl_data['encoding']:
						ctrl_data['encoding'][name] = len(ctrl_data['encoding'])
					self._add_statement(
						assigns = [ ctrl_data['signal'].eq(ctrl_data['encoding'][name]) ],
						domain  = ctrl_data['domain'],
						depth   = len(self._ctrl_stack)
					)
					return

		raise SyntaxError('`m.next = <...>` is only permitted inside an FSM state')

	def _pop_ctrl(self):
		name, data = self._ctrl_stack.pop()
		src_loc = data['src_loc']

		if name == 'If':
			if_tests, if_bodies = data['tests'], data['bodies']
			if_src_locs = data['src_locs']

			tests, cases = [], OrderedDict()
			for if_test, if_case in zip(if_tests + [ None ], if_bodies):
				if if_test is not None:
					if len(if_test) != 1:
						if_test = if_test.bool()
					tests.append(if_test)

				if if_test is not None:
					match = ('1' + '-' * (len(tests) - 1)).rjust(len(if_tests), '-')
				else:
					match = None
				cases[match] = if_case

			self._statements.append(Switch(Cat(tests), cases,
				src_loc = src_loc, case_src_locs = dict(zip(cases, if_src_locs))))

		if name == 'Switch':
			switch_test, switch_cases = data['test'], data['cases']
			switch_case_src_locs = data['case_src_locs']

			self._statements.append(Switch(switch_test, switch_cases,
				src_loc = src_loc, case_src_locs = switch_case_src_locs))

		if name == 'FSM':
			fsm_signal, fsm_reset, fsm_encoding, fsm_decoding, fsm_states = \
				data['signal'], data['reset'], data['encoding'], data['decoding'], data['states']
			fsm_state_src_locs = data['state_src_locs']
			if not fsm_states:
				return
			fsm_signal.width = bits_for(len(fsm_encoding) - 1)
			if fsm_reset is None:
				fsm_signal.reset = fsm_encoding[next(iter(fsm_states))]
			else:
				fsm_signal.reset = fsm_encoding[fsm_reset]
			# The FSM is encoded such that the state with encoding 0 is always the reset state.
			fsm_decoding.update((n, s) for s, n in fsm_encoding.items())
			fsm_signal.decoder = lambda n: "{}/{}".format(fsm_decoding[n], n)
			self._statements.append(Switch(fsm_signal,
				OrderedDict((fsm_encoding[name], stmts) for name, stmts in fsm_states.items()),
				src_loc = src_loc, case_src_locs = {fsm_encoding[name]: fsm_state_src_locs[name]
												for name in fsm_states}))

	def _add_statement(self, assigns, domain, depth, compat_mode = False):
		def domain_name(domain):
			if domain is None:
				return 'comb'
			else:
				return domain

		while len(self._ctrl_stack) > self.domain._depth:
			self._pop_ctrl()

		for stmt in Statement.cast(assigns):
			if not compat_mode and not isinstance(stmt, (Assign, Assert, Assume, Cover)):
				raise SyntaxError(f'Only assignments and property checks may be appended to d.{domain_name(domain)}')

			stmt._MustUse__used = True
			stmt = SampleDomainInjector(domain)(stmt)

			for signal in stmt._lhs_signals():
				if signal not in self._driving:
					self._driving[signal] = domain
				elif self._driving[signal] != domain:
					cd_curr = self._driving[signal]
					raise SyntaxError(
						f'Driver-driver conflict: trying to drive {signal!r} from d.{domain_name(domain)}, but it is '
						f'already driven from d.{domain_name(cd_curr)}')

			self._statements.append(stmt)

	def _add_submodule(self, submodule, name = None):
		if not hasattr(submodule, 'elaborate'):
			raise TypeError(f'Trying to add {submodule!r}, which does not implement .elaborate(), as a submodule')
		if name is None:
			self._anon_submodules.append(submodule)
		else:
			if name in self._named_submodules:
				raise NameError(f'Submodule named \'{name}\' already exists')
			self._named_submodules[name] = submodule

	def _get_submodule(self, name):
		if name in self._named_submodules:
			return self._named_submodules[name]
		else:
			raise AttributeError(f'No submodule named \'{name}\' exists')

	def _add_domain(self, cd):
		if cd.name in self._domains:
			raise NameError(f'Clock domain named \'{cd.name}\' already exists')
		self._domains[cd.name] = cd

	def _flush(self):
		while self._ctrl_stack:
			self._pop_ctrl()

	def elaborate(self, platform):
		self._flush()

		fragment = Fragment()
		for name in self._named_submodules:
			fragment.add_subfragment(Fragment.get(self._named_submodules[name], platform), name)
		for submodule in self._anon_submodules:
			fragment.add_subfragment(Fragment.get(submodule, platform), None)
		statements = SampleDomainInjector('sync')(self._statements)
		fragment.add_statements(statements)
		for signal, domain in self._driving.items():
			fragment.add_driver(signal, domain)
		fragment.add_domains(self._domains.values())
		fragment.generated.update(self._generated)
		return fragment



# SPDX-License-Identifier: BSD-2-Clause

''' First-in first-out queues. '''

from typing       import Union

from ..           import Elaboratable, Memory, Module, Mux, ResetSignal, Signal
from ..asserts    import Assert, Assume, Initial
from ..util.units import log2_int
from .cdc         import AsyncFFSynchronizer, FFSynchronizer
from .coding.gray import Decoder, Encoder

__all__ = (
	'AsyncFIFO',
	'AsyncFIFOBuffered',
	'FIFOInterface',
	'SyncFIFO',
	'SyncFIFOBuffered',
)


class FIFOInterface:
	_doc_template = '''
	{description}

	Parameters
	----------
	width : int
		Bit width of data entries.
	depth : int
		Depth of the queue. If zero, the FIFO cannot be read from or written to.
	{parameters}

	Attributes
	----------
	{attributes}
	w_data : Signal(width), in
		Input data.
	w_rdy : Signal(1), out
		Asserted if there is space in the queue, i.e. ``w_en`` can be asserted to write
		a new entry.
	w_en : Signal(1), in
		Write strobe. Latches ``w_data`` into the queue. Does nothing if ``w_rdy`` is not asserted.
	w_level : Signal(range(depth + 1)), out
		Number of unread entries.
	{w_attributes}
	r_data : Signal(width), out
		Output data. {r_data_valid}
	r_rdy : Signal(1), out
		Asserted if there is an entry in the queue, i.e. ``r_en`` can be asserted to read
		an existing entry.
	r_en : Signal(1), in
		Read strobe. Makes the next entry (if any) available on ``r_data`` at the next cycle.
		Does nothing if ``r_rdy`` is not asserted.
	r_level : Signal(range(depth + 1)), out
		Number of unread entries.
	{r_attributes}

	'''

	__doc__ = _doc_template.format(
		description = '''
		Data written to the input interface (``w_data``, ``w_rdy``, ``w_en``) is buffered and can be
		read at the output interface (``r_data``, ``r_rdy``, ``r_en``). The data entry written first
		to the input also appears first on the output.
		''',
		parameters = '''
		fwft : bool
			First-word fallthrough. If set, when ``r_rdy`` rises, the first entry is already
			available, i.e. ``r_data`` is valid. Otherwise, after ``r_rdy`` rises, it is necessary
			to strobe ``r_en`` for ``r_data`` to become valid.
		'''.strip(),
		r_data_valid = 'The conditions in which ``r_data`` is valid depends on the type of the queue.',
		attributes   = '',
		w_attributes = '',
		r_attributes = ''
	)

	def __init__(self, *, width: int, depth: int, fwft: bool) -> None:
		if not isinstance(width, int) or width < 0:
			raise TypeError(f'FIFO width must be a non-negative integer, not {width!r}')

		if not isinstance(depth, int) or depth < 0:
			raise TypeError(f'FIFO depth must be a non-negative integer, not {depth!r}')

		self.width = width
		self.depth = depth
		self.fwft  = fwft

		self.w_data = Signal(width, reset_less = True)
		self.w_rdy  = Signal() # writable; not full
		self.w_en   = Signal()
		self.w_level = Signal(range(depth + 1))

		self.r_data = Signal(width, reset_less = True)
		self.r_rdy  = Signal() # readable; not empty
		self.r_en   = Signal()
		self.r_level = Signal(range(depth + 1))


def _incr(signal, modulo: int) -> Union[Mux, int]:
	if modulo == 2 ** len(signal):
		return signal + 1
	else:
		return Mux(signal == modulo - 1, 0, signal + 1)


class SyncFIFO(Elaboratable, FIFOInterface):
	__doc__ = FIFOInterface._doc_template.format(
		description = '''
		Synchronous first in, first out queue.

		Read and write interfaces are accessed from the same clock domain. If different clock domains
		are needed, use :class:`AsyncFIFO`.
		'''.strip(),
		parameters = '''
		fwft : bool
			First-word fallthrough. If set, when the queue is empty and an entry is written into it,
			that entry becomes available on the output on the same clock cycle. Otherwise, it is
			necessary to assert ``r_en`` for ``r_data`` to become valid.
		'''.strip(),
		r_data_valid = "For FWFT queues, valid if ``r_rdy`` is asserted. "
					"For non-FWFT queues, valid on the next cycle after ``r_rdy`` and ``r_en`` have been asserted.",
		attributes = '''
		level : Signal(range(depth + 1)), out
			Number of unread entries. This level is the same between read and write for synchronous FIFOs.
		'''.strip(),
		r_attributes = '',
		w_attributes = ''
	)

	def __init__(self, *, width: int, depth: int, fwft: bool = True) -> None:
		super().__init__(width = width, depth = depth, fwft = fwft)

		self.level = Signal(range(depth + 1))

	def elaborate(self, platform) -> Module:
		m = Module()
		if self.depth == 0:
			m.d.comb += [
				self.w_rdy.eq(0),
				self.r_rdy.eq(0),
			]
			return m

		m.d.comb += [
			self.w_rdy.eq(self.level != self.depth),
			self.r_rdy.eq(self.level != 0),
			self.w_level.eq(self.level),
			self.r_level.eq(self.level),
		]

		do_read  = self.r_rdy & self.r_en
		do_write = self.w_rdy & self.w_en

		storage = Memory(width = self.width, depth = self.depth)
		w_port  = m.submodules.w_port = storage.write_port()
		r_port  = m.submodules.r_port = storage.read_port(
			domain = 'comb' if self.fwft else 'sync', transparent = self.fwft)
		produce = Signal(range(self.depth))
		consume = Signal(range(self.depth))

		m.d.comb += [
			w_port.addr.eq(produce),
			w_port.data.eq(self.w_data),
			w_port.en.eq(self.w_en & self.w_rdy),
		]
		with m.If(do_write):
			m.d.sync += produce.eq(_incr(produce, self.depth))

		m.d.comb += [
			r_port.addr.eq(consume),
			self.r_data.eq(r_port.data),
		]
		if not self.fwft:
			m.d.comb += r_port.en.eq(self.r_en)
		with m.If(do_read):
			m.d.sync += consume.eq(_incr(consume, self.depth))

		with m.If(do_write & ~do_read):
			m.d.sync += self.level.eq(self.level + 1)
		with m.If(do_read & ~do_write):
			m.d.sync += self.level.eq(self.level - 1)

		if platform == 'formal':
			# TODO: move this logic to SymbiYosys
			with m.If(Initial()):
				m.d.comb += [
					Assume(produce < self.depth),
					Assume(consume < self.depth),
				]
				with m.If(produce == consume):
					m.d.comb += Assume((self.level == 0) | (self.level == self.depth))
				with m.If(produce > consume):
					m.d.comb += Assume(self.level == (produce - consume))
				with m.If(produce < consume):
					m.d.comb += Assume(self.level == (self.depth + produce - consume))
			with m.Else():
				m.d.comb += [
					Assert(produce < self.depth),
					Assert(consume < self.depth),
				]
				with m.If(produce == consume):
					m.d.comb += Assert((self.level == 0) | (self.level == self.depth))
				with m.If(produce > consume):
					m.d.comb += Assert(self.level == (produce - consume))
				with m.If(produce < consume):
					m.d.comb += Assert(self.level == (self.depth + produce - consume))

		return m


class SyncFIFOBuffered(Elaboratable, FIFOInterface):
	__doc__ = FIFOInterface._doc_template.format(
		description = '''
		Buffered synchronous first in, first out queue.

		This queue's interface is identical to :class:`SyncFIFO` configured as ``fwft = True``, but it
		does not use asynchronous memory reads, which are incompatible with FPGA block RAMs.

		In exchange, the latency between an entry being written to an empty queue and that entry
		becoming available on the output is increased by one cycle compared to :class:`SyncFIFO`.
		'''.strip(),
		parameters = '''
		fwft : bool
			Always set.
		'''.strip(),
		attributes = '''
		level : Signal(range(depth + 1)), out
			Number of unread entries. This level is the same between read and write for synchronous FIFOs.
		'''.strip(),
		r_data_valid = 'Valid if ``r_rdy`` is asserted.',
		r_attributes = '',
		w_attributes = ''
	)

	def __init__(self, *, width: int, depth: int):
		super().__init__(width = width, depth = depth, fwft = True)

		self.level = Signal(range(depth + 1))

	def elaborate(self, platform) -> Module:
		m = Module()
		if self.depth == 0:
			m.d.comb += [
				self.w_rdy.eq(0),
				self.r_rdy.eq(0),
			]
			return m

		# Effectively, this queue treats the output register of the non-FWFT inner queue as
		# an additional storage element.
		m.submodules.unbuffered = fifo = SyncFIFO(
			width = self.width,
			depth = self.depth - 1,
			fwft = False
		)

		m.d.comb += [
			fifo.w_data.eq(self.w_data),
			fifo.w_en.eq(self.w_en),
			self.w_rdy.eq(fifo.w_rdy),
		]

		m.d.comb += [
			self.r_data.eq(fifo.r_data),
			fifo.r_en.eq(fifo.r_rdy & (~self.r_rdy | self.r_en)),
		]
		with m.If(fifo.r_en):
			m.d.sync += self.r_rdy.eq(1)
		with m.Elif(self.r_en):
			m.d.sync += self.r_rdy.eq(0)

		m.d.comb += [
			self.level.eq(fifo.level + self.r_rdy),
			self.w_level.eq(self.level),
			self.r_level.eq(self.level),
		]

		return m


class AsyncFIFO(Elaboratable, FIFOInterface):
	__doc__ = FIFOInterface._doc_template.format(
		description = '''
		Asynchronous first in, first out queue.

		Read and write interfaces are accessed from different clock domains, which can be set when
		constructing the FIFO.

		:class:`AsyncFIFO` can be reset from the write clock domain. When the write domain reset is
		asserted, the FIFO becomes empty. When the read domain is reset, data remains in the FIFO - the
		read domain logic should correctly handle this case.

		:class:`AsyncFIFO` only supports power of 2 depths. Unless ``exact_depth`` is specified,
		the ``depth`` parameter is rounded up to the next power of 2.
		'''.strip(),
		parameters = '''
		r_domain : str
			Read clock domain.
		w_domain : str
			Write clock domain.
		fwft : bool
			Always set.
		'''.strip(),
		attributes = '',
		r_data_valid = 'Valid if ``r_rdy`` is asserted.',
		r_attributes = '''
		r_rst : Signal(1), out
			Asserted, for at least one read-domain clock cycle, after the FIFO has been reset by
			the write-domain reset.
		'''.strip(),
		w_attributes = ''
	)

	def __init__(
		self, *, width: int, depth: int, r_domain: str = 'read', w_domain: str = 'write',
		exact_depth: bool = False
	) -> None:
		if depth != 0:
			try:
				depth_bits = log2_int(depth, need_pow2 = exact_depth)
				depth = 1 << depth_bits
			except ValueError:
				raise ValueError(
					f'AsyncFIFO only supports depths that are powers of 2; requested exact depth {depth} is not'
				) from None
		else:
			depth_bits = 0
		super().__init__(width = width, depth = depth, fwft = True)

		self.r_rst = Signal()
		self._r_domain = r_domain
		self._w_domain = w_domain
		self._ctr_bits = depth_bits + 1

	def elaborate(self, platform) -> Module:
		m = Module()
		if self.depth == 0:
			m.d.comb += [
				self.w_rdy.eq(0),
				self.r_rdy.eq(0),
			]
			return m

		# The design of this queue is the "style #2" from Clifford E. Cummings' paper "Simulation
		# and Synthesis Techniques for Asynchronous FIFO Design":
		# http://www.sunburst-design.com/papers/CummingsSNUG2002SJ_FIFO1.pdf

		do_write = self.w_rdy & self.w_en
		do_read  = self.r_rdy & self.r_en

		# TODO: extract this pattern into lib.cdc.GrayCounter
		produce_w_bin = Signal(self._ctr_bits)
		produce_w_nxt = Signal(self._ctr_bits)
		m.d.comb += produce_w_nxt.eq(produce_w_bin + do_write)
		m.d[self._w_domain] += produce_w_bin.eq(produce_w_nxt)

		# Note: Both read-domain counters must be reset_less (see comments below)
		consume_r_bin = Signal(self._ctr_bits, reset_less = True)
		consume_r_nxt = Signal(self._ctr_bits)
		m.d.comb += consume_r_nxt.eq(consume_r_bin + do_read)
		m.d[self._r_domain] += consume_r_bin.eq(consume_r_nxt)

		produce_w_gry = Signal(self._ctr_bits)
		produce_r_gry = Signal(self._ctr_bits)
		produce_enc = m.submodules.produce_enc = Encoder(self._ctr_bits)
		produce_cdc = m.submodules.produce_cdc = FFSynchronizer(  # noqa: F841
			produce_w_gry, produce_r_gry, o_domain = self._r_domain
		)
		m.d.comb += produce_enc.i.eq(produce_w_nxt),
		m.d[self._w_domain] += produce_w_gry.eq(produce_enc.o)

		consume_r_gry = Signal(self._ctr_bits, reset_less = True)
		consume_w_gry = Signal(self._ctr_bits)
		consume_enc = m.submodules.consume_enc = Encoder(self._ctr_bits)
		consume_cdc = m.submodules.consume_cdc = FFSynchronizer(  # noqa: F841
			consume_r_gry, consume_w_gry, o_domain = self._w_domain
		)
		m.d.comb += consume_enc.i.eq(consume_r_nxt)
		m.d[self._r_domain] += consume_r_gry.eq(consume_enc.o)

		consume_w_bin = Signal(self._ctr_bits)
		consume_dec = m.submodules.consume_dec = Decoder(self._ctr_bits)
		m.d.comb += consume_dec.i.eq(consume_w_gry),
		m.d[self._w_domain] += consume_w_bin.eq(consume_dec.o)

		produce_r_bin = Signal(self._ctr_bits)
		produce_dec = m.submodules.produce_dec = Decoder(self._ctr_bits)
		m.d.comb += produce_dec.i.eq(produce_r_gry),
		m.d.comb += produce_r_bin.eq(produce_dec.o)

		w_full  = Signal()
		r_empty = Signal()
		m.d.comb += [
			w_full.eq(
				(produce_w_gry[-1]  != consume_w_gry[-1]) &
				(produce_w_gry[-2]  != consume_w_gry[-2]) &
				(produce_w_gry[:-2] == consume_w_gry[:-2])
			),
			r_empty.eq(consume_r_gry == produce_r_gry),
		]

		m.d[self._w_domain] += self.w_level.eq((produce_w_bin - consume_w_bin))
		m.d.comb += self.r_level.eq((produce_r_bin - consume_r_bin))

		storage = Memory(width = self.width, depth = self.depth)
		w_port  = m.submodules.w_port = storage.write_port(domain = self._w_domain)
		r_port  = m.submodules.r_port = storage.read_port(
			domain = self._r_domain, transparent = False
		)
		m.d.comb += [
			w_port.addr.eq(produce_w_bin[:-1]),
			w_port.data.eq(self.w_data),
			w_port.en.eq(do_write),
			self.w_rdy.eq(~w_full),
		]
		m.d.comb += [
			r_port.addr.eq(consume_r_nxt[:-1]),
			self.r_data.eq(r_port.data),
			r_port.en.eq(1),
			self.r_rdy.eq(~r_empty),
		]

		# Reset handling to maintain FIFO and CDC invariants in the presence of a write-domain
		# reset.
		# There is a CDC hazard associated with resetting an async FIFO - Gray code counters which
		# are reset to 0 violate their Gray code invariant. One way to handle this is to ensure
		# that both sides of the FIFO are asynchronously reset by the same signal. We adopt a
		# slight variation on this approach - reset control rests entirely with the write domain.
		# The write domain's reset signal is used to asynchronously reset the read domain's
		# counters and force the FIFO to be empty when the write domain's reset is asserted.
		# This requires the two read domain counters to be marked as "reset_less", as they are
		# reset through another mechanism. See https://github.com/amaranth-lang/amaranth/issues/181
		# for the full discussion.
		w_rst = ResetSignal(domain = self._w_domain, allow_reset_less = True)
		r_rst = Signal()

		# Async-set-sync-release synchronizer avoids CDC hazards
		rst_cdc = m.submodules.rst_cdc = AsyncFFSynchronizer(w_rst, r_rst, o_domain = self._r_domain)  # noqa: F841

		# Decode Gray code counter synchronized from write domain to overwrite binary
		# counter in read domain.
		rst_dec = m.submodules.rst_dec = Decoder(self._ctr_bits)
		m.d.comb += rst_dec.i.eq(produce_r_gry)
		with m.If(r_rst):
			m.d.comb += r_empty.eq(1)
			m.d[self._r_domain] += consume_r_gry.eq(produce_r_gry)
			m.d[self._r_domain] += consume_r_bin.eq(rst_dec.o)
			m.d[self._r_domain] += self.r_rst.eq(1)
		with m.Else():
			m.d[self._r_domain] += self.r_rst.eq(0)

		if platform == 'formal':
			with m.If(Initial()):
				m.d.comb += Assume(produce_w_gry == (produce_w_bin ^ produce_w_bin[1:]))
				m.d.comb += Assume(consume_r_gry == (consume_r_bin ^ consume_r_bin[1:]))

		return m


class AsyncFIFOBuffered(Elaboratable, FIFOInterface):
	__doc__ = FIFOInterface._doc_template.format(
		description = '''
		Buffered asynchronous first in, first out queue.

		Read and write interfaces are accessed from different clock domains, which can be set when
		constructing the FIFO.

		:class:`AsyncFIFOBuffered` only supports power of 2 plus one depths. Unless ``exact_depth``
		is specified, the ``depth`` parameter is rounded up to the next power of 2 plus one.
		(The output buffer acts as an additional queue element.)

		This queue's interface is identical to :class:`AsyncFIFO`, but it has an additional register
		on the output, improving timing in case of block RAM that has large clock-to-output delay.

		In exchange, the latency between an entry being written to an empty queue and that entry
		becoming available on the output is increased by one cycle compared to :class:`AsyncFIFO`.
		'''.strip(),
		parameters = '''
		r_domain : str
			Read clock domain.
		w_domain : str
			Write clock domain.
		fwft : bool
			Always set.
		'''.strip(),
		attributes = '',
		r_data_valid = 'Valid if ``r_rdy`` is asserted.',
		r_attributes = '''
		r_rst : Signal(1), out
			Asserted, for at least one read-domain clock cycle, after the FIFO has been reset by
			the write-domain reset.
		'''.strip(),
		w_attributes = ''
	)

	def __init__(
		self, *, width: int, depth: int, r_domain: str = 'read', w_domain: str = 'write',
		exact_depth: bool = False
	) -> None:
		if depth != 0:
			try:
				depth_bits = log2_int(max(0, depth - 1), need_pow2 = exact_depth)
				depth = (1 << depth_bits) + 1
			except ValueError:
				raise ValueError(
					'AsyncFIFOBuffered only supports depths that are one higher than powers of 2; '
					f'requested exact depth {depth} is not'
				) from None
		super().__init__(width = width, depth = depth, fwft = True)

		self.r_rst = Signal()
		self._r_domain = r_domain
		self._w_domain = w_domain

	def elaborate(self, platform) -> Module:
		m = Module()
		if self.depth == 0:
			m.d.comb += [
				self.w_rdy.eq(0),
				self.r_rdy.eq(0),
			]
			return m

		m.submodules.unbuffered = fifo = AsyncFIFO(width = self.width, depth = self.depth - 1,
			r_domain = self._r_domain, w_domain = self._w_domain)

		m.d.comb += [
			fifo.w_data.eq(self.w_data),
			self.w_rdy.eq(fifo.w_rdy),
			fifo.w_en.eq(self.w_en),
		]

		r_consume_buffered = Signal()
		m.d.comb += r_consume_buffered.eq((self.r_rdy - self.r_en) & self.r_rdy)
		m.d[self._r_domain] += self.r_level.eq(fifo.r_level + r_consume_buffered)

		w_consume_buffered = Signal()
		m.submodules.consume_buffered_cdc = FFSynchronizer(
			r_consume_buffered, w_consume_buffered, o_domain = self._w_domain, stages = 4
		)
		m.d.comb += self.w_level.eq(fifo.w_level + w_consume_buffered)

		with m.If(self.r_en | ~self.r_rdy):
			m.d[self._r_domain] += [
				self.r_data.eq(fifo.r_data),
				self.r_rdy.eq(fifo.r_rdy),
				self.r_rst.eq(fifo.r_rst),
			]
			m.d.comb += [
				fifo.r_en.eq(1)
			]

		return m



# SPDX-License-Identifier: BSD-2-Clause

from warnings             import warn

from .lattice.machxo_2_3l import MachXO2Or3LPlatform as LatticeMachXO2Or3LPlatform
from .lattice.machxo_2_3l import MachXO2Platform     as LatticeMachXO2Platform
from .lattice.machxo_2_3l import MachXO3LPlatform    as LatticeMachXO3LPlatform


__all__ = (
	'LatticeMachXO2Or3LPlatform',
	'LatticeMachXO2Platform',
	'LatticeMachXO3LPlatform',
)

warn(
	'Instead of torii.platform.vendor.lattice_machxo_2_3l, use torii.platform.vendor.lattice.machxo_2_3l',
	DeprecationWarning, stacklevel = 2
)



# SPDX-License-Identifier: BSD-2-Clause

from collections.abc import Iterable
from linecache       import getlines
from re              import compile
from typing          import Union

__all__ = (
	'flatten',
	'get_linter_option',
	'get_linter_options',
	'union',
)

def flatten(i):
	for e in i:
		if isinstance(e, str) or not isinstance(e, Iterable):
			yield e
		else:
			yield from flatten(e)


def union(i, start = None):
	r = start
	for e in i:
		if r is None:
			r = e
		else:
			r |= e
	return r

def get_linter_options(filename: str) -> dict[str, Union[int, str]]:
	magic_comment = compile(r'^#\s*torii:\s*((?:\w+=\w+\s*)(?:,\s*\w+=\w+\s*)*)\n$')

	# Check the first five lines of the file, because it might not be first
	lines = getlines(filename)[0:5]
	if len(lines) > 0:
		matches = list(filter(lambda m: m is not None, map(magic_comment.match, lines)))

		if len(matches) > 0:
			return dict(map(lambda s: s.strip().split('=', 2), matches[0].group(1).split(',')))
	return dict()


def get_linter_option(
	filename: str , name: str, type: Union[type[bool], type[int]], default: Union[bool, int]
) -> Union[bool, int]:
	options = get_linter_options(filename)
	if name not in options:
		return default

	option = options[name]
	if type is bool:
		if option in ('1', 'yes', 'enable'):
			return True
		if option in ('0', 'no', 'disable'):
			return False
		return default
	if type is int:
		try:
			return int(option, 0)
		except ValueError:
			return default
	assert False



# SPDX-License-Identifier: BSD-2-Clause

from ... import Const, Elaboratable, Module, Signal

__all__ = (
	'Decoder',
	'Encoder',
)

class Encoder(Elaboratable):
	'''
	Encode binary to Gray code.

	Parameters
	----------
	width : int
		Bit width.

	Attributes
	----------
	i : Signal(width), in
		Natural binary input.
	o : Signal(width), out
		Encoded Gray code.

	'''

	def __init__(self, width: int) -> None:
		self.width = width

		self.i = Signal(width)
		self.o = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.comb += self.o.eq(self.i ^ self.i[1:])
		return m


class Decoder(Elaboratable):
	'''
	Decode Gray code to binary.

	Parameters
	----------
	width : int
		Bit width.

	Attributes
	----------
	i : Signal(width), in
		Gray code input.
	o : Signal(width), out
		Decoded natural binary.

	'''

	def __init__(self, width: int) -> None:
		self.width = width

		self.i = Signal(width)
		self.o = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		rhs = Const(0)
		for i in reversed(range(self.width)):
			rhs = rhs ^ self.i[i]
			m.d.comb += self.o[i].eq(rhs)
		return m



# SPDX-License-Identifier: BSD-2-Clause

from typing        import Union

from ..hdl         import ast, ir
from ..tools.yosys import YosysError, find_yosys
from .             import rtlil

__all__ = (
	'convert_fragment',
	'convert',
	'YosysError',
)


def _convert_rtlil_text(
	rtlil_text: str, *, strip_internal_attrs: bool = False, write_verilog_opts: tuple[str] = ()
) -> str:

	yosys = find_yosys()
	yosys_version = yosys.version()

	script = []
	script.append(f'read_rtlil <<rtlil\n{rtlil_text}\nrtlil')
	if yosys_version >= (0, 17):
		script.append('proc -nomux -norom')
	else:
		script.append('proc -nomux')
	script.append('memory_collect')

	if strip_internal_attrs:
		attr_map = []
		attr_map.append('-remove generator')
		attr_map.append('-remove top')
		attr_map.append('-remove src')
		attr_map.append('-remove torii.hierarchy')
		attr_map.append('-remove torii.decoding')
		script.append(f'attrmap {" ".join(attr_map)}')
		script.append(f'attrmap -modattr {" ".join(attr_map)}')

	script.append(f'write_verilog -norename {" ".join(write_verilog_opts)}')

	return yosys.run(['-q', '-'], '\n'.join(script),
		# At the moment, Yosys always shows a warning indicating that not all processes can be
		# translated to Verilog. We carefully emit only the processes that *can* be translated, and
		# squash this warning. Once Yosys' write_verilog pass is fixed, we should remove this.
		ignore_warnings = True
	)


def convert_fragment(*args, strip_internal_attrs: bool = False, **kwargs) -> tuple[str, ast.SignalDict]:
	rtlil_text, name_map = rtlil.convert_fragment(*args, **kwargs)
	return (_convert_rtlil_text(rtlil_text, strip_internal_attrs = strip_internal_attrs), name_map)


def convert(
	elaboratable: Union[ir.Fragment, ir.Elaboratable], name: str = 'top', platform = None, *, ports,
	emit_src: bool = True, strip_internal_attrs: bool = False, **kwargs
) -> str:

	fragment = ir.Fragment.get(elaboratable, platform).prepare(ports = ports, **kwargs)
	verilog_text, _ = convert_fragment(fragment, name, emit_src = emit_src, strip_internal_attrs = strip_internal_attrs)
	return verilog_text
